#!/usr/bin/env python
# cardinal_pythonlib/django/fields/restrictedcontentfile.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Django field class storing a file (by reference to a disk file, as for
"django.db.models.FileField") but also implementing limits on the maximum
upload size.**

"""

from typing import Any

# noinspection PyUnresolvedReferences
from django import forms
# noinspection PyUnresolvedReferences
from django.core.files.uploadedfile import UploadedFile
# noinspection PyUnresolvedReferences
from django.db import models
# noinspection PyUnresolvedReferences
from django.template.defaultfilters import filesizeformat
# noinspection PyUnresolvedReferences
from django.utils.translation import ugettext_lazy


# =============================================================================
# ContentTypeRestrictedFileField
# =============================================================================

class ContentTypeRestrictedFileField(models.FileField):
    """
    Same as ``FileField``, but you can specify:

    - ``content_types`` - list containing allowed content_types.
      Example: ``['application/pdf', 'image/jpeg']``
    - ``max_upload_size`` - a number indicating the maximum file size allowed
      for upload.

      .. code-block:: none

        2.5MB - 2621440
        5MB - 5242880
        10MB - 10485760
        20MB - 20971520
        50MB - 5242880
        100MB - 104857600
        250MB - 214958080
        500MB - 429916160

    See:

    - https://djangosnippets.org/snippets/2206/
    - https://docs.djangoproject.com/en/1.8/ref/files/uploads/
    - https://stackoverflow.com/questions/2472422/django-file-upload-size-limit
    """
    def __init__(self, *args, **kwargs) -> None:
        self.content_types = kwargs.pop("content_types", None)
        if self.content_types is None:
            self.content_types = []
        self.max_upload_size = kwargs.pop("max_upload_size", None)
        super().__init__(*args, **kwargs)

    def clean(self, *args, **kwargs) -> Any:
        data = super().clean(*args, **kwargs)
        # log.debug("data: {!r}", data)
        f = data.file
        if not isinstance(f, UploadedFile):  # RNC
            # no new file uploaded; there won't be a content-type to check
            return data
        # log.debug("f: {!r}", f)
        content_type = f.content_type
        if content_type not in self.content_types:
            raise forms.ValidationError(ugettext_lazy(
                'Filetype not supported.'))
        if hasattr(f, "size"):  # e.g. Django 2.1.2
            uploaded_file_size = f.size
        elif hasattr(f, "_size"):  # e.g. Django 1.8 ?
            # noinspection PyProtectedMember,PyUnresolvedReferences
            uploaded_file_size = f._size
        else:
            raise AssertionError(
                f"Don't know how to get file size from {f!r}")
        if (self.max_upload_size is not None and
                uploaded_file_size > self.max_upload_size):
            raise forms.ValidationError(ugettext_lazy(
                'Please keep filesize under %s. Current filesize %s')
                % (filesizeformat(self.max_upload_size),
                   filesizeformat(uploaded_file_size)))
        return data




#!/usr/bin/env python
# cardinal_pythonlib/athena_ohdsi.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to assist with the Athena OHDSI vocabularies.**

See https://athena.ohdsi.org/.

"""

import csv
import logging
from typing import Collection, Generator, Iterable, List

from cardinal_pythonlib.logs import BraceStyleAdapter
from cardinal_pythonlib.reprfunc import simple_repr
from cardinal_pythonlib.snomed import SnomedConcept

log = BraceStyleAdapter(logging.getLogger(__name__))


# =============================================================================
# Athena OHDSI mapping
# =============================================================================

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------

class AthenaVocabularyId(object):
    """
    Constant-holding class for Athena vocabulary IDs that we care about.
    """
    ICD10CM = "ICD10CM"
    ICD9CM = "ICD9CM"
    OPCS4 = "OPCS4"
    SNOMED = "SNOMED"


class AthenaRelationshipId(object):
    r"""
    Constant-holding class for Athena relationship IDs that we care about.
    To show all (there are lots!):

    .. code-block:: bash

        awk 'BEGIN {FS="\t"}; {print $3}' CONCEPT_RELATIONSHIP.csv | sort -u

    """
    IS_A = "Is a"  # "is a child of"
    MAPS_TO = "Maps to"  # converting between vocabularies
    MAPPED_FROM = "Mapped from"  # converting between vocabularies
    SUBSUMES = "Subsumes"  # "is a parent of"


# -----------------------------------------------------------------------------
# TSV row info classes
# -----------------------------------------------------------------------------

class AthenaConceptRow(object):
    """
    Simple information-holding class for ``CONCEPT.csv`` file from
    https://athena.ohdsi.org/ vocabulary download.
    """
    HEADER = [
        "concept_id", "concept_name", "domain_id", "vocabulary_id",
        "concept_class_id", "standard_concept", "concept_code",
        "valid_start_date", "valid_end_date", "invalid_reason"
    ]

    def __init__(self,
                 concept_id: str,
                 concept_name: str,
                 domain_id: str,
                 vocabulary_id: str,
                 concept_class_id: str,
                 standard_concept: str,
                 concept_code: str,
                 valid_start_date: str,
                 valid_end_date: str,
                 invalid_reason: str) -> None:
        """
        Argument order is important.

        Args:
            concept_id: Athena concept ID
            concept_name: Concept name in the originating system
            domain_id: e.g. "Observation", "Condition"
            vocabulary_id: e.g. "SNOMED", "ICD10CM"
            concept_class_id: e.g. "Substance", "3-char nonbill code"
            standard_concept: ?; e.g. "S"
            concept_code: concept code in the vocabulary (e.g. SNOMED-CT
                concept code like "3578611000001105" if vocabulary_id is
                "SNOMED"; ICD-10 code like "F32.2" if vocabulary_is is
                "ICD10CM"; etc.)
            valid_start_date: date in YYYYMMDD format
            valid_end_date: date in YYYYMMDD format
            invalid_reason: ? (but one can guess)
        """
        self.concept_id = int(concept_id)
        self.concept_name = concept_name
        self.domain_id = domain_id
        self.vocabulary_id = vocabulary_id
        self.concept_class_id = concept_class_id
        self.standard_concept = standard_concept
        self.concept_code = concept_code
        self.valid_start_date = valid_start_date
        self.valid_end_date = valid_end_date
        self.invalid_reason = invalid_reason
        # self.sort_context_concept_to_match = None

    def __repr__(self) -> str:
        return simple_repr(self, self.HEADER)

    def __str__(self) -> str:
        return (
            f"Vocabulary {self.vocabulary_id}, concept {self.concept_code} "
            f"({self.concept_name}) -> Athena concept {self.concept_id}"
        )

    # I looked at sorting them to find the best. Not wise; would need human
    # review. Just use all valid codes.

    _ = '''

    def set_sort_context_concept_to_match(self,
                                          concept: "AthenaConceptRow") -> None:
        self.sort_context_concept_to_match = concept

    def __lt__(self, other: "AthenaConceptRow") -> bool:
        """
        Compares using "less than" being equivalent to "preferable to".

        So, returns True if "self" is better than other, and False if "self" is
        worse than other; that is, all tests look like "return self is better
        than other".

        BINNED. We will use human judgement.
        """
        invalid_s = bool(self.invalid_reason)
        invalid_o = bool(other.invalid_reason)
        if invalid_s != invalid_o:
            # better not to have an "invalid" reason;
            # empty strings are "less than" full ones
            return invalid_s < invalid_o
        if self.valid_end_date != other.valid_end_date:
            # better to have a later end date
            return self.valid_end_date > other.valid_end_date
        if self.valid_start_date != other.valid_start_date:
            # better to have an earlier start date
            return self.valid_start_date < other.valid_end_date
        if self.sort_context_concept_to_match:
            # Which is closer to our target context?
            c = self.sort_context_concept_to_match
            sp = self.match_tuple(c)
            op = other.match_tuple(c)
            log.info(
                "Tie-breaking to match {c}: {s} ({sp} points) vs "
                "{o} ({op} points)",
                s=self, sp=sp, o=other, op=op, c=c
            )
            # More matching points is better
            return self.match_tuple(c) > other.match_tuple(c)
        log.warning("Tie-breaking {} and {} by ID", self, other)
        # Arbitrarily, better to have an earlier (lower) concept ID.
        return self.concept_id < other.concept_id

    def match_tuple(self, target: "AthenaConceptRow") -> Tuple[float, float]:
        """
        Returns a score reflecting our similarity to the target.

        See

        - https://stackoverflow.com/questions/8897593/similarity-between-two-text-documents
        - https://stackoverflow.com/questions/2380394/simple-implementation-of-n-gram-tf-idf-and-cosine-similarity-in-python
        - https://spacy.io/usage/vectors-similarity -- data not included
        - https://radimrehurek.com/gensim/index.html
        - https://radimrehurek.com/gensim/tut3.html
        - https://scikit-learn.org/stable/
        - https://www.nltk.org/

        BINNED. We will use human judgement.
        """  # noqa
        self_words = set(x.lower() for x in self.concept_name.split())
        other_words = set(x.lower() for x in target.concept_name.split())
        # More matching words better
        n_matching_words = len(self_words & other_words)
        # More words better (often more specific)
        n_words = len(self_words)
        return float(n_matching_words), float(n_words)

    '''

    def snomed_concept(self) -> SnomedConcept:
        """
        Assuming this Athena concept reflects a SnomedConcept, returns it.

        (Asserts if it isn't.)
        """
        assert self.vocabulary_id == AthenaVocabularyId.SNOMED
        return SnomedConcept(int(self.concept_code), self.concept_name)


class AthenaConceptRelationshipRow(object):
    """
    Simple information-holding class for ``CONCEPT_RELATIONSHIP.csv`` file from
    https://athena.ohdsi.org/ vocabulary download.
    """
    HEADER = [
        "concept_id_1", "concept_id_2", "relationship_id",
        "valid_start_date", "valid_end_date", "invalid_reason",
    ]

    def __init__(self,
                 concept_id_1: str,
                 concept_id_2: str,
                 relationship_id: str,
                 valid_start_date: str,
                 valid_end_date: str,
                 invalid_reason: str) -> None:
        """
        Argument order is important.

        Args:
            concept_id_1: Athena concept ID #1
            concept_id_2: Athena concept ID #2
            relationship_id: e.g. "Is a", "Has legal category"
            valid_start_date: date in YYYYMMDD format
            valid_end_date: date in YYYYMMDD format
            invalid_reason: ? (but one can guess)
        """
        self.concept_id_1 = int(concept_id_1)
        self.concept_id_2 = int(concept_id_2)
        self.relationship_id = relationship_id
        self.valid_start_date = valid_start_date
        self.valid_end_date = valid_end_date
        self.invalid_reason = invalid_reason

    def __repr__(self) -> str:
        return simple_repr(self, self.HEADER)

    def __str__(self) -> str:
        return (
            f"Athena concept relationship {self.concept_id_1} "
            f"{self.relationship_id!r} {self.concept_id_2}"
        )


# -----------------------------------------------------------------------------
# Fetch data from TSV files
# -----------------------------------------------------------------------------

# noinspection DuplicatedCode
def get_athena_concepts(
        tsv_filename: str = "",
        cached_concepts: Iterable[AthenaConceptRow] = None,
        vocabulary_ids: Collection[str] = None,
        concept_codes: Collection[str] = None,
        concept_ids: Collection[int] = None,
        not_vocabulary_ids: Collection[str] = None,
        not_concept_codes: Collection[str] = None,
        not_concept_ids: Collection[int] = None,
        encoding: str = "utf-8") -> List[AthenaConceptRow]:
    """
    From the Athena ``CONCEPT.csv`` tab-separated value file, return a list
    of concepts matching the restriction criteria.

    Args:
        tsv_filename:
            filename
        cached_concepts:
            alternative to tsv_filename
        vocabulary_ids:
            permissible ``vocabulary_id`` values, or None or an empty list for
            all
        concept_codes:
            permissible ``concept_code`` values, or None or an empty list for
            all
        concept_ids:
            permissible ``concept_id`` values, or None or an empty list for all
        not_vocabulary_ids:
            impermissible ``vocabulary_id`` values, or None or an empty list
            for none
        not_concept_codes:
            impermissible ``concept_code`` values, or None or an empty list for
            none
        not_concept_ids:
            impermissible ``concept_id`` values, or None or an empty list for
            none
        encoding:
            encoding for input files

    Returns:
        list: of :class:`AthenaConceptRow` objects

    Test and timing code:

    .. code-block:: python

        import logging
        import timeit
        logging.basicConfig(level=logging.DEBUG)

        from cardinal_pythonlib.athena_ohdsi import (
            get_athena_concepts,
            get_athena_concept_relationships,
        )

        concept_filename = "CONCEPT.csv"
        cr_filename = "CONCEPT_RELATIONSHIP.csv"
        testcode = "175898006"
        testid = 46067884

        concept_testcode = '''
        get_athena_concepts(concept_filename, concept_codes=[testcode])
        '''
        cr_testcode = '''
        get_athena_concept_relationships(cr_filename, concept_id_1_values=[testid])
        '''

        timeit.timeit(cr_testcode, number=1, globals=globals())
        # Initial method: 33.6 s (for 9.9m rows on a Windows laptop).
        # Chain of generators: 21.5 s. Better.

        timeit.timeit(concept_testcode, number=1, globals=globals())
        # After speedup: 3.9 s for 1.1m rows.

    """  # noqa
    assert bool(tsv_filename) != bool(cached_concepts), (
        "Specify either tsv_filename or cached_concepts"
    )
    n_rows_read = 0

    def gen_rows() -> Generator[AthenaConceptRow, None, None]:
        nonlocal n_rows_read
        with open(tsv_filename, 'r', encoding=encoding) as tsvin:
            reader = csv.reader(tsvin, delimiter="\t")
            header = next(reader, None)
            if header != AthenaConceptRow.HEADER:
                raise ValueError(
                    f"Athena concept file has unexpected header: {header!r}; "
                    f"expected {AthenaConceptRow.HEADER!r}")
            for row in reader:
                n_rows_read += 1
                concept = AthenaConceptRow(*row)
                yield concept

    def filter_vocab(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.vocabulary_id in vocabulary_ids:
                yield concept

    def filter_code(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.concept_code in concept_codes:
                yield concept

    def filter_id(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.concept_id in concept_ids:
                yield concept

    def filter_not_vocab(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.vocabulary_id not in not_vocabulary_ids:
                yield concept

    def filter_not_code(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.concept_code not in not_concept_codes:
                yield concept

    def filter_not_id(concepts_: Iterable[AthenaConceptRow]) \
            -> Generator[AthenaConceptRow, None, None]:
        for concept in concepts_:
            if concept.concept_id not in not_concept_ids:
                yield concept

    # Build up the fastest pipeline we can.
    if tsv_filename:
        log.info(f"Loading Athena concepts from file: {tsv_filename}")
        gen = gen_rows()
    else:
        log.info("Using cached Athena concepts")
        gen = cached_concepts
    # Positive checks
    if vocabulary_ids:
        gen = filter_vocab(gen)
    if concept_codes:
        gen = filter_code(gen)
    if concept_ids:
        gen = filter_id(gen)
    # Negative checks
    if not_vocabulary_ids:
        gen = filter_not_vocab(gen)
    if not_concept_codes:
        gen = filter_not_code(gen)
    if not_concept_ids:
        gen = filter_not_id(gen)

    concepts = list(concept for concept in gen)
    log.debug(f"Retrieved {len(concepts)} concepts from {n_rows_read} rows")
    return concepts


# noinspection DuplicatedCode
def get_athena_concept_relationships(
        tsv_filename: str = "",
        cached_concept_relationships: Iterable[AthenaConceptRelationshipRow] = None,  # noqa
        concept_id_1_values: Collection[int] = None,
        concept_id_2_values: Collection[int] = None,
        relationship_id_values: Collection[str] = None,
        not_concept_id_1_values: Collection[int] = None,
        not_concept_id_2_values: Collection[int] = None,
        not_relationship_id_values: Collection[str] = None,
        encoding: str = "utf-8") \
        -> List[AthenaConceptRelationshipRow]:
    """
    From the Athena ``CONCEPT_RELATIONSHIP.csv`` tab-separated value file,
    return a list of relationships matching the restriction criteria.

    Args:
        tsv_filename:
            filename
        cached_concept_relationships:
            alternative to tsv_filename
        concept_id_1_values:
            permissible ``concept_id_1`` values, or None or an empty list for
            all
        concept_id_2_values:
            permissible ``concept_id_2`` values, or None or an empty list for
            all
        relationship_id_values:
            permissible ``relationship_id`` values, or None or an empty list
            for all
        not_concept_id_1_values:
            impermissible ``concept_id_1`` values, or None or an empty list for
            none
        not_concept_id_2_values:
            impermissible ``concept_id_2`` values, or None or an empty list for
            none
        not_relationship_id_values:
            impermissible ``relationship_id`` values, or None or an empty list
            for none
        encoding:
            encoding for input files

    Returns:
        list: of :class:`AthenaConceptRelationshipRow` objects

    """
    assert bool(tsv_filename) != bool(cached_concept_relationships), (
        "Specify either tsv_filename or cached_concept_relationships"
    )
    n_rows_read = 0

    def gen_rows() -> Generator[AthenaConceptRelationshipRow, None, None]:
        nonlocal n_rows_read
        with open(tsv_filename, 'r', encoding=encoding) as tsvin:
            reader = csv.reader(tsvin, delimiter="\t")
            header = next(reader, None)
            if header != AthenaConceptRelationshipRow.HEADER:
                raise ValueError(
                    f"Athena concept relationship file has unexpected header: "
                    f"{header!r}; expected "
                    f"{AthenaConceptRelationshipRow.HEADER!r}")
            for row in reader:
                n_rows_read += 1
                rel = AthenaConceptRelationshipRow(*row)
                yield rel

    def filter_rel(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.relationship_id in relationship_id_values:
                yield rel

    def filter_c1(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.concept_id_1 in concept_id_1_values:
                yield rel

    def filter_c2(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.concept_id_2 in concept_id_2_values:
                yield rel

    def filter_not_rel(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.relationship_id not in not_relationship_id_values:
                yield rel

    def filter_not_c1(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.concept_id_1 not in not_concept_id_1_values:
                yield rel

    def filter_not_c2(rels: Iterable[AthenaConceptRelationshipRow]) \
            -> Generator[AthenaConceptRelationshipRow, None, None]:
        for rel in rels:
            if rel.concept_id_2 not in not_concept_id_2_values:
                yield rel

    # Build up the fastest pipeline we can.
    if tsv_filename:
        log.info(f"Loading Athena concept relationships from file: "
                 f"{tsv_filename}")
        gen = gen_rows()
    else:
        log.info("Using cached Athena concept relationships")
        gen = cached_concept_relationships
    # Positive checks
    if relationship_id_values:
        gen = filter_rel(gen)
    if concept_id_1_values:
        gen = filter_c1(gen)
    if concept_id_2_values:
        gen = filter_c2(gen)
    # Negative checks
    if not_relationship_id_values:
        gen = filter_not_rel(gen)
    if not_concept_id_1_values:
        gen = filter_not_c1(gen)
    if not_concept_id_2_values:
        gen = filter_not_c2(gen)

    relationships = list(rel for rel in gen)
    log.debug(f"Retrieved {len(relationships)} relationships from "
              f"{n_rows_read} rows")
    return relationships



#!/usr/bin/env python
# cardinal_pythonlib/dicts.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Dictionary manipulations.**

"""

from typing import Any, Callable, Dict, Hashable, List, Optional


# =============================================================================
# Dictionaries
# =============================================================================

def get_case_insensitive_dict_key(d: Dict, k: str) -> Optional[str]:
    """
    Within the dictionary ``d``, find a key that matches (in case-insensitive
    fashion) the key ``k``, and return it (or ``None`` if there isn't one).
    """
    for key in d.keys():
        if k.lower() == key.lower():
            return key
    return None


def merge_dicts(*dict_args: Dict) -> Dict:
    """
    Given any number of dicts, shallow-copy them and merge into a new dict.
    Precedence goes to key/value pairs in dicts that are later in the list.

    See https://stackoverflow.com/questions/38987.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def merge_two_dicts(x: Dict, y: Dict) -> Dict:
    """
    Given two dicts, merge them into a new dict as a shallow copy, e.g.

    .. code-block:: python

        z = merge_two_dicts(x, y)

    If you can guarantee Python 3.5, then a simpler syntax is:

    .. code-block:: python

        z = {**x, **y}

    See https://stackoverflow.com/questions/38987.
    """
    z = x.copy()
    z.update(y)
    return z


def rename_key(d: Dict[str, Any], old: str, new: str) -> None:
    """
    Rename a key in dictionary ``d`` from ``old`` to ``new``, in place.
    """
    d[new] = d.pop(old)


def rename_keys(d: Dict[str, Any], mapping: Dict[str, str]) -> Dict[str, Any]:
    """
    Returns a copy of the dictionary ``d`` with its keys renamed according to
    ``mapping``.

    Args:
        d: the starting dictionary
        mapping: a dictionary of the format ``{old_key_name: new_key_name}``

    Returns:
        a new dictionary

    Keys that are not in ``mapping`` are left unchanged.
    The input parameters are not modified.
    """
    result = {}  # type: Dict[str, Any]
    for k, v in d.items():
        if k in mapping:
            k = mapping[k]
        result[k] = v
    return result


def rename_keys_in_dict(d: Dict[str, Any], renames: Dict[str, str]) -> None:
    """
    Renames, IN PLACE, the keys in ``d`` according to the mapping in
    ``renames``.
    
    Args:
        d: a dictionary to modify 
        renames: a dictionary of the format ``{old_key_name: new_key_name}``
        
    See
    https://stackoverflow.com/questions/4406501/change-the-name-of-a-key-in-dictionary.
    """  # noqa
    for old_key, new_key in renames.items():
        if new_key == old_key:
            continue
        if old_key in d:
            if new_key in d:
                raise ValueError(
                    f"rename_keys_in_dict: renaming {old_key!r} -> "
                    f"{new_key!r} but new key already exists")
            d[new_key] = d.pop(old_key)


def prefix_dict_keys(d: Dict[str, Any], prefix: str) -> Dict[str, Any]:
    """
    Returns a dictionary that's a copy of as ``d`` but with ``prefix``
    prepended to its keys.
    """
    result = {}  # type: Dict[str, Any]
    for k, v in d.items():
        result[prefix + k] = v
    return result


def reversedict(d: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Takes a ``k -> v`` mapping and returns a ``v -> k`` mapping.
    """
    return {v: k for k, v in d.items()}


def set_null_values_in_dict(d: Dict[str, Any],
                            null_literals: List[Any]) -> None:
    """
    Within ``d`` (in place), replace any values found in ``null_literals`` with
    ``None``.
    """
    if not null_literals:
        return
    # DO NOT add/delete values to/from a dictionary during iteration, but it
    # is OK to modify existing keys:
    #       https://stackoverflow.com/questions/6777485
    #       https://stackoverflow.com/questions/2315520
    #       https://docs.python.org/3/library/stdtypes.html#dict-views
    for k, v in d.items():
        if v in null_literals:
            d[k] = None


# noinspection PyPep8
def map_keys_to_values(keys: List[Any], d: Dict[Any, Any], default: Any = None,
                       raise_if_missing: bool = False,
                       omit_if_missing: bool = False) -> List[Any]:
    """
    The ``d`` dictionary contains a ``key -> value`` mapping.

    We start with a list of potential keys in ``keys``, and return a list of
    corresponding values -- substituting ``default`` if any are missing,
    or raising :exc:`KeyError` if ``raise_if_missing`` is true, or omitting the
    entry if ``omit_if_missing`` is true.
    """
    result = []
    for k in keys:
        if raise_if_missing and k not in d:
            raise ValueError("Missing key: " + repr(k))
        if omit_if_missing and k not in d:
            continue
        result.append(d.get(k, default))
    return result


def dict_diff(d1: Dict[Any, Any], d2: Dict[Any, Any],
              deleted_value: Any = None) -> Dict[Any, Any]:
    """
    Returns a representation of the changes that need to be made to ``d1`` to
    create ``d2``.

    Args:
        d1: a dictionary
        d2: another dictionary
        deleted_value: value to use for deleted keys; see below

    Returns:
        dict: a dictionary of the format ``{k: v}`` where the ``k``/``v`` pairs
        are key/value pairs that are absent from ``d1`` and present in ``d2``,
        or present in both but with different values (in which case the ``d2``
        value is shown). If a key ``k`` is present in ``d1`` but absent in
        ``d2``, the result dictionary has the entry ``{k: deleted_value}``.

    """
    changes = {k: v for k, v in d2.items()
               if k not in d1 or d2[k] != d1[k]}
    for k in d1.keys():
        if k not in d2:
            changes[k] = deleted_value
    return changes


def delete_keys(d: Dict[Any, Any],
                keys_to_delete: List[Any],
                keys_to_keep: List[Any]) -> None:
    """
    Deletes keys from a dictionary, in place.

    Args:
        d:
            dictonary to modify
        keys_to_delete:
            if any keys are present in this list, they are deleted...
        keys_to_keep:
            ... unless they are present in this list.
    """
    for k in keys_to_delete:
        if k in d and k not in keys_to_keep:
            del d[k]


# =============================================================================
# Lazy dictionaries
# =============================================================================

class LazyDict(dict):
    """
    A dictionary that only evaluates the argument to :func:`setdefault` or
    :func:`get` if it needs to.
    
    See
    https://stackoverflow.com/questions/17532929/how-to-implement-a-lazy-setdefault.
    
    The ``*args``/``**kwargs`` parts are useful, but we don't want to have to
    name 'thunk' explicitly.
    """  # noqa
    def get(self, key: Hashable, thunk: Any = None,
            *args: Any, **kwargs: Any) -> Any:
        if key in self:
            return self[key]
        elif callable(thunk):
            return thunk(*args, **kwargs)
        else:
            return thunk

    def setdefault(self, key: Hashable, thunk: Any = None,
                   *args: Any, **kwargs: Any) -> Any:
        if key in self:
            return self[key]
        elif callable(thunk):
            return dict.setdefault(self, key, thunk(*args, **kwargs))
        else:
            return dict.setdefault(self, key, thunk)


class LazyButHonestDict(dict):
    """
    A dictionary that provides alternatives to :func:`get` and
    :func:`setdefault`, namely :func:`lazyget` and :func:`lazysetdefault`,
    that only evaluate their arguments if they have to.

    See
    https://stackoverflow.com/questions/17532929/how-to-implement-a-lazy-setdefault.

    Compared to the StackOverflow version: no obvious need to have a default
    returning ``None``, when we're implementing this as a special function.
    In contrast, helpful to have ``*args``/``**kwargs`` options.
    """  # noqa
    def lazyget(self, key: Hashable, thunk: Callable,
                *args: Any, **kwargs: Any) -> Any:
        if key in self:
            return self[key]
        else:
            return thunk(*args, **kwargs)

    def lazysetdefault(self, key: Hashable, thunk: Callable,
                       *args: Any, **kwargs: Any) -> Any:
        if key in self:
            return self[key]
        else:
            return self.setdefault(key, thunk(*args, **kwargs))


# =============================================================================
# HashableDict
# =============================================================================

class HashableDict(dict):
    """
    A dictionary that can be hashed.

    See https://stackoverflow.com/questions/1151658/python-hashable-dicts.
    """
    def __hash__(self) -> int:
        return hash(tuple(sorted(self.items())))


# =============================================================================
# CaseInsensitiveDict
# =============================================================================

class CaseInsensitiveDict(dict):
    """
    A case-insensitive dictionary, as per
    https://stackoverflow.com/questions/2082152/case-insensitive-dictionary/32888599#32888599,
    with updates for Python 3 and type hinting.
    
    See also
    
    - https://docs.python.org/3/tutorial/datastructures.html#dictionaries
    - https://docs.python.org/3/library/stdtypes.html#mapping-types-dict
    
    Test code:
    
    .. code-block:: python
    
        from cardinal_pythonlib.dicts import CaseInsensitiveDict
        
        d1 = CaseInsensitiveDict()  # d1 is now: {}
        d2 = CaseInsensitiveDict({'A': 1, 'b': 2})  # d2 is now: {'a': 1, 'b': 2}
        d3 = CaseInsensitiveDict(C=3, d=4)  # d3 is now: {'c': 3, 'd': 4}
        
        d1.update({'E': 5, 'f': 6})  # d1 is now: {'e': 5, 'f': 6}
        d1.update(G=7, h=8)  # d1 is now: {'e': 5, 'f': 6, 'g': 7, 'h': 8}
        'H' in d1  # True
        d1['I'] = 9  # None, and key 'i' added
        del d1['I']  # None, and key 'i' deleted
        d1.pop('H')  # 8
        d1.get('E')  # 5
        d1.get('Z')  # None
        d1.setdefault('J', 10)  # 10, and key 'j' added
        d1.update([('K', 11), ('L', 12)])
        d1  # {'e': 5, 'f': 6, 'g': 7, 'j': 10, 'k': 11, 'l': 12}

    """  # noqa

    @classmethod
    def _k(cls, key: Any) -> Any:
        """
        Convert key to lower case, if it's a string.
        """
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs) -> None:
        """
        Dictionary initialization.

        - Optional positional argument is ``mapping`` or ``iterable``. If an
          iterable, its elements are iterables of length 2.
          (Note that passing ``None`` is different from not passing anything,
          hence the signature. The type of the first argument, if present, is
          ``Union[Mapping, Iterable[Tuple[Any, Any]]]``.)
        - Keyword arguments are key/value pairs.
        """
        super().__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key: Any) -> Any:
        """
        Given a key, return the associated value. Implements ``d[key]`` as an
        rvalue.
        """
        return super().__getitem__(self.__class__._k(key))

    def __setitem__(self, key: Any, value: Any) -> None:
        """
        Sets the value for a key. Implements ``d[key] = value``.
        """
        super().__setitem__(self.__class__._k(key), value)

    def __delitem__(self, key: Any) -> None:
        """
        Deletes the item with the specified key. Implements ``del d[key]``.
        Raises :exc:`KeyError` if absent.
        """
        super().__delitem__(self.__class__._k(key))

    def __contains__(self, key: Any) -> bool:
        """
        Is the key in the dictionary? Implements ``key in d``.
        """
        return super().__contains__(self.__class__._k(key))

    # has_key() was removed in Python 3.0
    # https://docs.python.org/3.1/whatsnew/3.0.html#builtins

    def pop(self, key: Any, *args, **kwargs) -> Any:
        """
        Retrieves/returns the item and removes it. Takes a single optional
        argument, being the default to return if the key is not present
        (otherwise raises :exc:`KeyError`). Note that supplying a default of
        ``None`` is different to supplying no default.
        """
        return super().pop(self.__class__._k(key), *args, **kwargs)

    def get(self, key: Any, default: Any = None) -> Any:
        """
        If the key is in the dictionary, return the corresponding value;
        otherwise, return ``default``, which defaults to ``None``.
        """
        return super().get(self.__class__._k(key), default)

    def setdefault(self, key: Any, default: Any = None) -> Any:
        """
        As per the Python docs:

        If ``key`` is in the dictionary, return its value. If not, insert
        ``key`` with a value of ``default`` and return ``default``. ``default``
        defaults to ``None``.
        """
        return super().setdefault(self.__class__._k(key), default)

    def update(self, *args, **kwargs) -> None:
        """
        As per the Python docs:

        Update the dictionary with the key/value pairs from ``other``,
        overwriting existing keys. Return ``None``.

        :func:`update``accepts either another dictionary object or an iterable
        of key/value pairs (as tuples or other iterables of length two). If
        keyword arguments are specified, the dictionary is then updated with
        those key/value pairs: ``d.update(red=1, blue=2)``.

        ... so the type of the first argument, if present, is ``Union[Mapping,
        .Iterable[Tuple[Any, Any]]]``.
        """
        # noinspection PyTypeChecker
        super().update(self.__class__(*args, **kwargs))

    def _convert_keys(self) -> None:
        """
        Ensure all our keys are in lower case.
        """
        for k in list(self.keys()):
            v = super().pop(k)
            self.__setitem__(k, v)



#!/usr/bin/env python
# cardinal_pythonlib/logs.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for logging.**

See https://docs.python.org/3.4/howto/logging.html#library-config

USER CODE should use the following general methods.

(a) Simple:

    .. code-block:: python

        import logging
        log = logging.getLogger(__name__)  # for your own logs
        logging.basicConfig()

(b) More complex:

    .. code-block:: python

        import logging
        log = logging.getLogger(__name__)
        logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATEFMT,
                            level=loglevel)

(c) Using colour conveniently:

    .. code-block:: python

        import logging
        mylogger = logging.getLogger(__name__)
        rootlogger = logging.getLogger()

        from whisker.log import configure_logger_for_colour
        configure_logger_for_colour(rootlogger)


LIBRARY CODE should use the following general methods.

.. code-block:: python

    import logging
    log = logging.getLogger(__name__)

    # ... and if you want to suppress output unless the user configures logs:
    log.addHandler(logging.NullHandler())
    # ... which only needs to be done in the __init__.py for the package
    #     https://stackoverflow.com/questions/12296214

    # LIBRARY CODE SHOULD NOT ADD ANY OTHER HANDLERS; see above.

DO NOT call this module "logging"! Many things may get confused.

"""

from html import escape
from inspect import Parameter, signature
import json
import logging
import os
from typing import Any, Callable, Dict, List, Optional, TextIO, Tuple, Union

from colorlog import ColoredFormatter

# =============================================================================
# Quick configuration of a specific log format
# =============================================================================

LOG_FORMAT = '%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:%(message)s'
LOG_FORMAT_WITH_PID = (
    f'%(asctime)s.%(msecs)03d:%(levelname)s:{os.getpid()}:%(name)s:%(message)s'
)

LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'

LOG_COLORS = {'DEBUG': 'cyan',
              'INFO': 'green',
              'WARNING': 'bold_yellow',
              'ERROR': 'bold_red',
              'CRITICAL': 'bold_white,bg_red'}


def get_monochrome_handler(
        extranames: List[str] = None,
        with_process_id: bool = False,
        with_thread_id: bool = False,
        stream: TextIO = None) -> logging.StreamHandler:
    """
    Gets a monochrome log handler using a standard format.

    Args:
        extranames: additional names to append to the logger's name
        with_process_id: include the process ID in the logger's name?
        with_thread_id: include the thread ID in the logger's name?
        stream: ``TextIO`` stream to send log output to

    Returns:
        the :class:`logging.StreamHandler`

    """
    fmt = "%(asctime)s.%(msecs)03d"
    if with_process_id or with_thread_id:
        procinfo = []  # type: List[str]
        if with_process_id:
            procinfo.append("p%(process)d")
        if with_thread_id:
            procinfo.append("t%(thread)d")
        fmt += f" [{'.'.join(procinfo)}]"
    extras = ":" + ":".join(extranames) if extranames else ""
    fmt += f" %(name)s{extras}:%(levelname)s: "
    fmt += "%(message)s"
    f = logging.Formatter(fmt, datefmt=LOG_DATEFMT, style='%')
    h = logging.StreamHandler(stream)
    h.setFormatter(f)
    return h


def get_colour_handler(extranames: List[str] = None,
                       with_process_id: bool = False,
                       with_thread_id: bool = False,
                       stream: TextIO = None) -> logging.StreamHandler:
    """
    Gets a colour log handler using a standard format.

    Args:
        extranames: additional names to append to the logger's name
        with_process_id: include the process ID in the logger's name?
        with_thread_id: include the thread ID in the logger's name?
        stream: ``TextIO`` stream to send log output to

    Returns:
        the :class:`logging.StreamHandler`

    """
    fmt = "%(white)s%(asctime)s.%(msecs)03d"  # this is dim white = grey
    if with_process_id or with_thread_id:
        procinfo = []  # type: List[str]
        if with_process_id:
            procinfo.append("p%(process)d")
        if with_thread_id:
            procinfo.append("t%(thread)d")
        fmt += f" [{'.'.join(procinfo)}]"
    extras = ":" + ":".join(extranames) if extranames else ""
    fmt += f" %(name)s{extras}:%(levelname)s: "
    fmt += "%(reset)s%(log_color)s%(message)s"
    cf = ColoredFormatter(fmt,
                          datefmt=LOG_DATEFMT,
                          reset=True,
                          log_colors=LOG_COLORS,
                          secondary_log_colors={},
                          style='%')
    ch = logging.StreamHandler(stream)
    ch.setFormatter(cf)
    return ch


def configure_logger_for_colour(logger: logging.Logger,
                                level: int = logging.INFO,
                                remove_existing: bool = False,
                                extranames: List[str] = None,
                                with_process_id: bool = False,
                                with_thread_id: bool = False) -> None:
    """
    Applies a preconfigured datetime/colour scheme to a logger.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Args:
        logger: logger to modify
        level: log level to set
        remove_existing: remove existing handlers from logger first?
        extranames: additional names to append to the logger's name
        with_process_id: include the process ID in the logger's name?
        with_thread_id: include the thread ID in the logger's name?
    """
    if remove_existing:
        logger.handlers = []  # https://stackoverflow.com/questions/7484454
    handler = get_colour_handler(extranames,
                                 with_process_id=with_process_id,
                                 with_thread_id=with_thread_id)
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(level)


def main_only_quicksetup_rootlogger(level: int = logging.DEBUG,
                                    with_process_id: bool = False,
                                    with_thread_id: bool = False) -> None:
    """
    Quick function to set up the root logger for colour.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Args:
        level: log level to set
        with_process_id: include the process ID in the logger's name?
        with_thread_id: include the thread ID in the logger's name?
    """
    # Nasty. Only call from "if __name__ == '__main__'" clauses!
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, level, remove_existing=True,
                                with_process_id=with_process_id,
                                with_thread_id=with_thread_id)
    # logging.basicConfig(level=level)


# =============================================================================
# Generic log functions
# =============================================================================

def remove_all_logger_handlers(logger: logging.Logger) -> None:
    """
    Remove all handlers from a logger.

    Args:
        logger: logger to modify
    """
    while logger.handlers:
        h = logger.handlers[0]
        logger.removeHandler(h)


def reset_logformat(logger: logging.Logger,
                    fmt: str,
                    datefmt: str = '%Y-%m-%d %H:%M:%S') -> None:
    """
    Create a new formatter and apply it to the logger.

    :func:`logging.basicConfig` won't reset the formatter if another module
    has called it, so always set the formatter like this.

    Args:
        logger: logger to modify
        fmt: passed to the ``fmt=`` argument of :class:`logging.Formatter`
        datefmt: passed to the ``datefmt=`` argument of
            :class:`logging.Formatter`
    """
    handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)
    handler.setFormatter(formatter)
    remove_all_logger_handlers(logger)
    logger.addHandler(handler)
    logger.propagate = False


def reset_logformat_timestamped(logger: logging.Logger,
                                extraname: str = "",
                                level: int = logging.INFO) -> None:
    """
    Apply a simple time-stamped log format to an existing logger, and set
    its loglevel to either ``logging.DEBUG`` or ``logging.INFO``.

    Args:
        logger: logger to modify
        extraname: additional name to append to the logger's name
        level: log level to set
    """
    namebit = extraname + ":" if extraname else ""
    fmt = ("%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:" + namebit +
           "%(message)s")
    # logger.info(fmt)
    reset_logformat(logger, fmt=fmt)
    # logger.info(fmt)
    logger.setLevel(level)


# =============================================================================
# Helper functions
# =============================================================================

def configure_all_loggers_for_colour(remove_existing: bool = True) -> None:
    """
    Applies a preconfigured datetime/colour scheme to ALL logger.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Generally MORE SENSIBLE just to apply a handler to the root logger.

    Args:
        remove_existing: remove existing handlers from logger first?

    """
    handler = get_colour_handler()
    apply_handler_to_all_logs(handler, remove_existing=remove_existing)


def apply_handler_to_root_log(handler: logging.Handler,
                              remove_existing: bool = False) -> None:
    """
    Applies a handler to all logs, optionally removing existing handlers.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Generally MORE SENSIBLE just to apply a handler to the root logger.

    Args:
        handler: the handler to apply
        remove_existing: remove existing handlers from logger first?
    """
    rootlog = logging.getLogger()
    if remove_existing:
        rootlog.handlers = []
    rootlog.addHandler(handler)


def apply_handler_to_all_logs(handler: logging.Handler,
                              remove_existing: bool = False) -> None:
    """
    Applies a handler to all logs, optionally removing existing handlers.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Generally MORE SENSIBLE just to apply a handler to the root logger.

    Args:
        handler: the handler to apply
        remove_existing: remove existing handlers from logger first?
    """
    # noinspection PyUnresolvedReferences
    for name, obj in logging.Logger.manager.loggerDict.items():
        if remove_existing:
            obj.handlers = []  # https://stackoverflow.com/questions/7484454
        obj.addHandler(handler)


def copy_root_log_to_file(filename: str,
                          fmt: str = LOG_FORMAT,
                          datefmt: str = LOG_DATEFMT) -> None:
    """
    Copy all currently configured logs to the specified file.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.
    """
    fh = logging.FileHandler(filename)
    # default file mode is 'a' for append
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)
    fh.setFormatter(formatter)
    apply_handler_to_root_log(fh)


def copy_all_logs_to_file(filename: str,
                          fmt: str = LOG_FORMAT,
                          datefmt: str = LOG_DATEFMT) -> None:
    """
    Copy all currently configured logs to the specified file.

    Should ONLY be called from the ``if __name__ == 'main'`` script;
    see https://docs.python.org/3.4/howto/logging.html#library-config.

    Args:
        filename: file to send log output to
        fmt: passed to the ``fmt=`` argument of :class:`logging.Formatter`
        datefmt: passed to the ``datefmt=`` argument of
            :class:`logging.Formatter`
    """
    fh = logging.FileHandler(filename)
    # default file mode is 'a' for append
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)
    fh.setFormatter(formatter)
    apply_handler_to_all_logs(fh)


# noinspection PyProtectedMember
def get_formatter_report(f: logging.Formatter) -> Optional[Dict[str, str]]:
    """
    Returns information on a log formatter, as a dictionary.
    For debugging.
    """
    if f is None:
        return None
    return {
        '_fmt': f._fmt,
        'datefmt': f.datefmt,
        '_style': str(f._style),
    }


def get_handler_report(h: logging.Handler) -> Dict[str, Any]:
    """
    Returns information on a log handler, as a dictionary.
    For debugging.
    """
    # noinspection PyUnresolvedReferences
    return {
        'get_name()': h.get_name(),
        'level': h.level,
        'formatter': get_formatter_report(h.formatter),
        'filters': h.filters,
    }


def get_log_report(log: Union[logging.Logger,
                              logging.PlaceHolder]) -> Dict[str, Any]:
    """
    Returns information on a log, as a dictionary. For debugging.
    """
    if isinstance(log, logging.Logger):
        # suppress invalid error for Logger.manager:
        # noinspection PyUnresolvedReferences
        return {
            '(object)': str(log),
            'level': log.level,
            'disabled': log.disabled,
            'propagate': log.propagate,
            'parent': str(log.parent),
            'manager': str(log.manager),
            'handlers': [get_handler_report(h) for h in log.handlers],
        }
    elif isinstance(log, logging.PlaceHolder):
        return {
            "(object)": str(log),
        }
    else:
        raise ValueError(f"Unknown object type: {log!r}")


def print_report_on_all_logs() -> None:
    """
    Use :func:`print` to report information on all logs.
    """
    d = {}
    # noinspection PyUnresolvedReferences
    for name, obj in logging.Logger.manager.loggerDict.items():
        d[name] = get_log_report(obj)
    rootlogger = logging.getLogger()
    d['(root logger)'] = get_log_report(rootlogger)
    print(json.dumps(d, sort_keys=True, indent=4, separators=(',', ': ')))


def set_level_for_logger_and_its_handlers(log: logging.Logger,
                                          level: int) -> None:
    """
    Set a log level for a log and all its handlers.

    Args:
        log: log to modify
        level: log level to set
    """
    log.setLevel(level)
    for h in log.handlers:  # type: logging.Handler
        h.setLevel(level)


# =============================================================================
# HTML formatter
# =============================================================================

class HtmlColorFormatter(logging.Formatter):
    """
    Class to format Python logs in coloured HTML.
    """
    log_colors = {
        logging.DEBUG: '#008B8B',  # dark cyan
        logging.INFO: '#00FF00',  # green
        logging.WARNING: '#FFFF00',  # yellow
        logging.ERROR: '#FF0000',  # red
        logging.CRITICAL: '#FF0000',  # red
    }
    log_background_colors = {
        logging.DEBUG: None,
        logging.INFO: None,
        logging.WARNING: None,
        logging.ERROR: None,
        logging.CRITICAL: '#FFFFFF',  # white
    }

    def __init__(self, append_br: bool = False,
                 replace_nl_with_br: bool = True) -> None:
        r"""
        Args:
            append_br: append ``<br>`` to each line?
            replace_nl_with_br: replace ``\n`` with ``<br>`` in messages?

        See https://hg.python.org/cpython/file/3.5/Lib/logging/__init__.py
        """
        super().__init__(
            fmt='%(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            style='%'
        )
        self.append_br = append_br
        self.replace_nl_with_br = replace_nl_with_br

    def format(self, record: logging.LogRecord) -> str:
        """
        Internal function to format the :class:`LogRecord` as HTML.

        See https://docs.python.org/3.4/library/logging.html#logging.LogRecord
        """

        # message = super().format(record)
        super().format(record)
        # Since fmt does not contain asctime, the Formatter.format()
        # will not write asctime (since its usesTime()) function will be
        # false. Therefore:
        record.asctime = self.formatTime(record, self.datefmt)
        bg_col = self.log_background_colors[record.levelno]
        msg = escape(record.getMessage())
        # escape() won't replace \n but will replace & etc.
        if self.replace_nl_with_br:
            msg = msg.replace("\n", "<br>")
        html = (
            '<span style="color:#008B8B">{time}.{ms:03d} {name}:{lvname}: '
            '</span><span style="color:{color}{bg}">{msg}</font>{br}'.format(
                time=record.asctime,
                ms=int(record.msecs),
                name=record.name,
                lvname=record.levelname,
                color=self.log_colors[record.levelno],
                msg=msg,
                bg=f";background-color:{bg_col}" if bg_col else "",
                br="<br>" if self.append_br else "",
            )
        )
        # print("record.__dict__: {}".format(record.__dict__))
        # print("html: {}".format(html))
        return html


# =============================================================================
# HtmlColorHandler
# =============================================================================

class HtmlColorHandler(logging.StreamHandler):
    """
    HTML handler (using :class:`HtmlColorFormatter`) that sends output to a
    function, e.g. for display in a Qt window
    """
    def __init__(self, logfunction: Callable[[str], None],
                 level: int = logging.INFO) -> None:
        super().__init__()
        self.logfunction = logfunction
        self.setFormatter(HtmlColorFormatter())
        self.setLevel(level)

    def emit(self, record: logging.LogRecord) -> None:
        """
        Internal function to process a :class:`LogRecord`.
        """
        # noinspection PyBroadException
        try:
            html = self.format(record)
            self.logfunction(html)
        except:  # nopep8
            self.handleError(record)


# =============================================================================
# Brace formatters, for log.info("{}, {}", "hello", "world")
# =============================================================================

# - https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles  # noqa
# - https://stackoverflow.com/questions/13131400/logging-variable-data-with-new-format-string  # noqa
# - https://stackoverflow.com/questions/13131400/logging-variable-data-with-new-format-string/24683360#24683360  # noqa
# ... plus modifications to use inspect.signature() not inspect.getargspec()
# ... plus a performance tweak so we're not calling signature() every time
# See also:
# - https://www.simonmweber.com/2014/11/24/python-logging-traps.html

class BraceMessage(object):
    """
    Class to represent a message that includes a message including braces
    (``{}``) and a set of ``args``/``kwargs``. When converted to a ``str``,
    the message is realized via ``msg.format(*args, **kwargs)``.
    """
    def __init__(self,
                 fmt: str,
                 args: Tuple[Any, ...],
                 kwargs: Dict[str, Any]) -> None:
        # This version uses args and kwargs, not *args and **kwargs, for
        # performance reasons:
        # https://stackoverflow.com/questions/31992424/performance-implications-of-unpacking-dictionaries-in-python  # noqa
        # ... and since we control creation entirely, we may as well go fast
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs
        # print("Creating BraceMessage with: fmt={}, args={}, "
        #       "kwargs={}".format(repr(fmt), repr(args), repr(kwargs)))

    def __str__(self) -> str:
        return self.fmt.format(*self.args, **self.kwargs)


class BraceStyleAdapter(logging.LoggerAdapter):
    def __init__(self,
                 logger: logging.Logger,
                 pass_special_logger_args: bool = True,
                 strip_special_logger_args_from_fmt: bool = False) -> None:
        """
        Wraps a logger so we can use ``{}``-style string formatting.

        Args:
            logger:
                a logger
            pass_special_logger_args:
                should we continue to pass any special arguments to the logger
                itself? True is standard; False probably brings a slight
                performance benefit, but prevents log.exception() from working
                properly, as the 'exc_info' parameter will be stripped.
            strip_special_logger_args_from_fmt:
                If we're passing special arguments to the logger, should we
                remove them from the argments passed to the string formatter?
                There is no obvious cost to saying no.

        Specimen use:

        .. code-block:: python

            import logging
            from cardinal_pythonlib.logs import BraceStyleAdapter, main_only_quicksetup_rootlogger

            log = BraceStyleAdapter(logging.getLogger(__name__))

            main_only_quicksetup_rootlogger(level=logging.DEBUG)

            log.info("Hello {}, {title} {surname}!", "world", title="Mr", surname="Smith")
            # 2018-09-17 16:13:50.404 __main__:INFO: Hello world, Mr Smith!

        """  # noqa
        # noinspection PyTypeChecker
        super().__init__(logger=logger, extra=None)
        self.pass_special_logger_args = pass_special_logger_args
        self.strip_special_logger_args_from_fmt = strip_special_logger_args_from_fmt  # noqa
        # getargspec() returns:
        #   named tuple: ArgSpec(args, varargs, keywords, defaults)
        #   ... args = list of parameter names
        #   ... varargs = names of the * parameters, or None
        #   ... keywords = names of the ** parameters, or None
        #   ... defaults = tuple of default argument values, or None
        # signature() returns a Signature object:
        #   ... parameters: ordered mapping of name -> Parameter
        #   ... ... https://docs.python.org/3/library/inspect.html#inspect.Parameter  # noqa
        # Direct equivalence:
        #   https://github.com/praw-dev/praw/issues/541
        # So, old:
        # logargnames = getargspec(self.logger._log).args[1:]
        # and new:
        # noinspection PyProtectedMember
        sig = signature(self.logger._log)
        self.logargnames = [p.name for p in sig.parameters.values()
                            if p.kind == Parameter.POSITIONAL_OR_KEYWORD]
        # e.g.: ['level', 'msg', 'args', 'exc_info', 'extra', 'stack_info']
        # print("self.logargnames: " + repr(self.logargnames))

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level):
            # print("log: msg={}, args={}, kwargs={}".format(
            #     repr(msg), repr(args), repr(kwargs)))
            if self.pass_special_logger_args:
                msg, log_kwargs = self.process(msg, kwargs)
                # print("... log: msg={}, log_kwargs={}".format(
                #     repr(msg), repr(log_kwargs)))
            else:
                log_kwargs = {}
            # noinspection PyProtectedMember
            self.logger._log(level, BraceMessage(msg, args, kwargs), (),
                             **log_kwargs)

    def process(self, msg: str,
                kwargs: Dict[str, Any]) -> Tuple[str, Dict[str, Any]]:
        special_param_names = [k for k in kwargs.keys()
                               if k in self.logargnames]
        log_kwargs = {k: kwargs[k] for k in special_param_names}
        # ... also: remove them from the starting kwargs?
        if self.strip_special_logger_args_from_fmt:
            for k in special_param_names:
                kwargs.pop(k)
        return msg, log_kwargs


def get_log_with_null_handler(name: str) -> logging.Logger:
    """
    For use by library functions. Returns a log with the specifed name that
    has a null handler attached, and a :class:`BraceStyleAdapter`.
    """
    log = logging.getLogger(name)
    log.addHandler(logging.NullHandler())
    return log


def get_brace_style_log_with_null_handler(name: str) -> BraceStyleAdapter:
    """
    For use by library functions. Returns a log with the specifed name that
    has a null handler attached, and a :class:`BraceStyleAdapter`.
    """
    log = logging.getLogger(name)
    log.addHandler(logging.NullHandler())
    return BraceStyleAdapter(log)


# =============================================================================
# Testing
# =============================================================================

if __name__ == '__main__':
    """
    Command-line validation checks.
    """
    main_only_quicksetup_rootlogger(logging.INFO)
    _log = BraceStyleAdapter(logging.getLogger(__name__))
    _log.info("1. Hello!")
    _log.info("1. Hello, {}!", "world")
    _log.info("1. Hello, foo={foo}, bar={bar}!", foo="foo", bar="bar")
    _log.info("1. Hello, {}; foo={foo}, bar={bar}!", "world", foo="foo",
              bar="bar")
    _log.info("1. Hello, {}; foo={foo}, bar={bar}!", "world", foo="foo",
              bar="bar", extra={'somekey': 'somevalue'})



#!/usr/bin/env python
# cardinal_pythonlib/colander_utils.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions for working with colander.**

Colander: https://docs.pylonsproject.org/projects/colander/en/latest/

"""

import random
from typing import (Any, Callable, Dict, Iterable, List, Optional,
                    Tuple, TYPE_CHECKING, Union)

from cardinal_pythonlib.datetimefunc import (
    coerce_to_pendulum,
    PotentialDatetimeType,
)
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
# noinspection PyUnresolvedReferences
import colander
# noinspection PyUnresolvedReferences
from colander import (
    Boolean,
    Date,
    DateTime,  # NB name clash with pendulum
    Email,
    Integer,
    Invalid,
    Length,
    MappingSchema,
    SchemaNode,
    SchemaType,
    String,
)
from deform.widget import (
    CheckboxWidget,
    DateTimeInputWidget,
    HiddenWidget,
)
from pendulum import DateTime as Pendulum  # NB name clash with colander
from pendulum.parsing.exceptions import ParserError

if TYPE_CHECKING:
    # noinspection PyProtectedMember,PyUnresolvedReferences
    from colander import _SchemaNode

log = get_brace_style_log_with_null_handler(__name__)

ColanderNullType = type(colander.null)
ValidatorType = Callable[[SchemaNode, Any], None]  # called as v(node, value)

# =============================================================================
# Debugging options
# =============================================================================

DEBUG_DANGER_VALIDATION = False

if DEBUG_DANGER_VALIDATION:
    log.warning("Debugging options enabled!")

# =============================================================================
# Constants
# =============================================================================

EMAIL_ADDRESS_MAX_LEN = 255  # https://en.wikipedia.org/wiki/Email_address
SERIALIZED_NONE = ""  # has to be a string; avoid "None" like the plague!


# =============================================================================
# New generic SchemaType classes
# =============================================================================

class PendulumType(SchemaType):
    """
    Colander :class:`SchemaType` for :class:`Pendulum` date/time objects.
    """
    def __init__(self, use_local_tz: bool = True):
        self.use_local_tz = use_local_tz
        super().__init__()  # not necessary; SchemaType has no __init__

    def serialize(self,
                  node: SchemaNode,
                  appstruct: Union[PotentialDatetimeType,
                                   ColanderNullType]) \
            -> Union[str, ColanderNullType]:
        """
        Serializes Python object to string representation.
        """
        if not appstruct:
            return colander.null
        try:
            appstruct = coerce_to_pendulum(appstruct,
                                           assume_local=self.use_local_tz)
        except (ValueError, ParserError) as e:
            raise Invalid(
                node,
                f"{appstruct!r} is not a pendulum.DateTime object; "
                f"error was {e!r}")
        return appstruct.isoformat()

    def deserialize(self,
                    node: SchemaNode,
                    cstruct: Union[str, ColanderNullType]) \
            -> Optional[Pendulum]:
        """
        Deserializes string representation to Python object.
        """
        if not cstruct:
            return colander.null
        try:
            result = coerce_to_pendulum(cstruct,
                                        assume_local=self.use_local_tz)
        except (ValueError, ParserError) as e:
            raise Invalid(node,
                          f"Invalid date/time: value={cstruct!r}, error={e!r}")
        return result


class AllowNoneType(SchemaType):
    """
    Serializes ``None`` to ``''``, and deserializes ``''`` to ``None``;
    otherwise defers to the parent type.

    A type which accepts serializing ``None`` to ``''`` and deserializing
    ``''`` to ``None``. When the value is not equal to ``None``/``''``, it will
    use (de)serialization of the given type. This can be used to make nodes
    optional.

    Example:

    .. code-block:: python

        date = colander.SchemaNode(
            colander.NoneType(colander.DateTime()),
            default=None,
            missing=None,
        )

    NOTE ALSO that Colander nodes explicitly never validate a missing value;
    see ``colander/__init__.py``, in :func:`_SchemaNode.deserialize`. We want
    them to do so, essentially so we can pass in ``None`` to a form but have
    the form refuse to validate if it's still ``None`` at submission.

    """
    def __init__(self, type_: SchemaType) -> None:
        self.type_ = type_

    def serialize(self, node: SchemaNode,
                  value: Any) -> Union[str, ColanderNullType]:
        """
        Serializes Python object to string representation.
        """
        if value is None:
            retval = ''
        else:
            # noinspection PyUnresolvedReferences
            retval = self.type_.serialize(node, value)
        # log.debug("AllowNoneType.serialize: {!r} -> {!r}", value, retval)
        return retval

    def deserialize(self, node: SchemaNode,
                    value: Union[str, ColanderNullType]) -> Any:
        """
        Deserializes string representation to Python object.
        """
        if value is None or value == '':
            retval = None
        else:
            # noinspection PyUnresolvedReferences
            retval = self.type_.deserialize(node, value)
        # log.debug("AllowNoneType.deserialize: {!r} -> {!r}", value, retval)
        return retval


# =============================================================================
# Node helper functions
# =============================================================================

def get_values_and_permissible(values: Iterable[Tuple[Any, str]],
                               add_none: bool = False,
                               none_description: str = "[None]") \
        -> Tuple[List[Tuple[Any, str]], List[Any]]:
    """
    Used when building Colander nodes.

    Args:
        values: an iterable of tuples like ``(value, description)`` used in
            HTML forms

        add_none: add a tuple ``(None, none_description)`` at the start of
            ``values`` in the result?

        none_description: the description used for ``None`` if ``add_none``
            is set

    Returns:
        a tuple ``(values, permissible_values)``, where

        - ``values`` is what was passed in (perhaps with the addition of the
          "None" tuple at the start)
        - ``permissible_values`` is a list of all the ``value`` elements of
          the original ``values``

    """
    permissible_values = list(x[0] for x in values)
    # ... does not include the None value; those do not go to the validator
    if add_none:
        none_tuple = (SERIALIZED_NONE, none_description)
        values = [none_tuple] + list(values)
    return values, permissible_values


def get_child_node(parent: "_SchemaNode", child_name: str) -> "_SchemaNode":
    """
    Returns a child node from an instantiated :class:`colander.SchemaNode`
    object. Such nodes are not accessible via ``self.mychild`` but must be
    accessed via ``self.children``, which is a list of child nodes.

    Args:
        parent: the parent node object
        child_name: the name of the child node

    Returns:
        the child node

    Raises:
        :exc:`StopIteration` if there isn't one
    """
    return next(c for c in parent.children if c.name == child_name)


# =============================================================================
# Validators
# =============================================================================

class EmailValidatorWithLengthConstraint(Email):
    """
    The Colander ``Email`` validator doesn't check length. This does.
    """
    def __init__(self, *args, min_length: int = 0, **kwargs) -> None:
        self._length = Length(min_length, EMAIL_ADDRESS_MAX_LEN)
        super().__init__(*args, **kwargs)

    def __call__(self, node: SchemaNode, value: Any) -> None:
        self._length(node, value)
        super().__call__(node, value)  # call Email regex validator


# =============================================================================
# Other new generic SchemaNode classes
# =============================================================================
# Note that we must pass both *args and **kwargs upwards, because SchemaNode
# does some odd stuff with clone().

# -----------------------------------------------------------------------------
# Simple types
# -----------------------------------------------------------------------------

class OptionalIntNode(SchemaNode):
    """
    Colander node accepting integers but also blank values (i.e. it's
    optional).
    """
    # YOU CANNOT USE ARGUMENTS THAT INFLUENCE THE STRUCTURE, because these Node
    # objects get default-copied by Deform.
    @staticmethod
    def schema_type() -> SchemaType:
        return AllowNoneType(Integer())

    default = None
    missing = None


class OptionalStringNode(SchemaNode):
    """
    Colander node accepting strings but allowing them to be blank (optional).

    Coerces None to ``""`` when serializing; otherwise it is coerced to
    ``"None"``, i.e. a string literal containing the word "None", which is much
    more wrong.
    """
    @staticmethod
    def schema_type() -> SchemaType:
        return AllowNoneType(String(allow_empty=True))

    default = ""
    missing = ""


class MandatoryStringNode(SchemaNode):
    """
    Colander string node, where the string is obligatory.

    CAVEAT: WHEN YOU PASS DATA INTO THE FORM, YOU MUST USE

    .. code-block:: python

        appstruct = {
            somekey: somevalue or "",
            #                  ^^^^^
            #                  without this, None is converted to "None"
        }
    """
    @staticmethod
    def schema_type() -> SchemaType:
        return String(allow_empty=False)


class HiddenIntegerNode(OptionalIntNode):
    """
    Colander node containing an integer, that is hidden to the user.
    """
    widget = HiddenWidget()


class HiddenStringNode(OptionalStringNode):
    """
    Colander node containing an optional string, that is hidden to the user.
    """
    widget = HiddenWidget()


class BooleanNode(SchemaNode):
    """
    Colander node representing a boolean value with a checkbox widget.
    """
    schema_type = Boolean
    widget = CheckboxWidget()

    def __init__(self, *args, title: str = "?", label: str = "",
                 default: bool = False, **kwargs) -> None:
        self.title = title  # above the checkbox
        self.label = label or title  # to the right of the checkbox
        self.default = default
        self.missing = default
        super().__init__(*args, **kwargs)


# -----------------------------------------------------------------------------
# Email addresses
# -----------------------------------------------------------------------------

class OptionalEmailNode(OptionalStringNode):
    """
    Colander string node, where the string can be blank but if not then it
    must look like a valid e-mail address.
    """
    validator = EmailValidatorWithLengthConstraint()


class MandatoryEmailNode(MandatoryStringNode):
    """
    Colander string node, requiring something that looks like a valid e-mail
    address.
    """
    validator = EmailValidatorWithLengthConstraint()


# -----------------------------------------------------------------------------
# Date/time types
# -----------------------------------------------------------------------------

class DateTimeSelectorNode(SchemaNode):
    """
    Colander node containing a date/time.
    """
    schema_type = DateTime
    missing = None


class DateSelectorNode(SchemaNode):
    """
    Colander node containing a date.
    """
    schema_type = Date
    missing = None


DEFAULT_WIDGET_DATE_OPTIONS_FOR_PENDULUM = dict(
    # http://amsul.ca/pickadate.js/date/#formatting-rules
    format='yyyy-mm-dd',
    selectMonths=True,
    selectYears=True,
)
DEFAULT_WIDGET_TIME_OPTIONS_FOR_PENDULUM = dict(
    # See http://amsul.ca/pickadate.js/time/#formatting-rules
    # format='h:i A',  # the default, e.g. "11:30 PM"
    format='HH:i',  # e.g. "23:30"
    interval=30,
)


class OptionalPendulumNodeLocalTZ(SchemaNode):
    """
    Colander node containing an optional :class:`Pendulum` date/time, in which
    the date/time is assumed to be in the local timezone.
    """
    @staticmethod
    def schema_type() -> SchemaType:
        return AllowNoneType(PendulumType(use_local_tz=True))

    default = None
    missing = None
    widget = DateTimeInputWidget(
        date_options=DEFAULT_WIDGET_DATE_OPTIONS_FOR_PENDULUM,
        time_options=DEFAULT_WIDGET_TIME_OPTIONS_FOR_PENDULUM,
    )


OptionalPendulumNode = OptionalPendulumNodeLocalTZ  # synonym for back-compatibility  # noqa


class OptionalPendulumNodeUTC(SchemaNode):
    """
    Colander node containing an optional :class:`Pendulum` date/time, in which
    the date/time is assumed to be UTC.
    """
    @staticmethod
    def schema_type() -> SchemaType:
        return AllowNoneType(PendulumType(use_local_tz=False))

    default = None
    missing = None
    widget = DateTimeInputWidget(
        date_options=DEFAULT_WIDGET_DATE_OPTIONS_FOR_PENDULUM,
        time_options=DEFAULT_WIDGET_TIME_OPTIONS_FOR_PENDULUM,
    )


# -----------------------------------------------------------------------------
# Safety-checking nodes
# -----------------------------------------------------------------------------

class ValidateDangerousOperationNode(MappingSchema):
    """
    Colander node that can be added to forms allowing dangerous operations
    (e.g. deletion of data). The node shows the user a code and requires the
    user to type that code in, before it will permit the form to proceed.

    For this to work, the containing form *must* inherit from
    :class:`DynamicDescriptionsForm` with ``dynamic_descriptions=True``.

    Usage is simple, like this:

    .. code-block:: python

        class AddSpecialNoteSchema(CSRFSchema):
            table_name = HiddenStringNode()
            server_pk = HiddenIntegerNode()
            note = MandatoryStringNode(widget=TextAreaWidget(rows=20, cols=80))
            danger = ValidateDangerousOperationNode()

    """
    target = HiddenStringNode()
    user_entry = MandatoryStringNode(title="Validate this dangerous operation")

    def __init__(self, *args, length: int = 4, allowed_chars: str = None,
                 **kwargs) -> None:
        """
        Args:
            length: code length required from the user
            allowed_chars: string containing the permitted characters
                (by default, digits)
        """
        self.allowed_chars = allowed_chars or "0123456789"
        self.length = length
        super().__init__(*args, **kwargs)

    # noinspection PyUnusedLocal
    def after_bind(self, node: SchemaNode, kw: Dict[str, Any]) -> None:
        # Accessing the nodes is fiddly!
        target_node = get_child_node(self, "target")
        # Also, this whole thing is a bit hard to get your head around.
        # - This function will be called every time the form is accessed.
        # - The first time (fresh form load), there will be no value in
        #   "target", so we set "target.default", and "target" will pick up
        #   that default value.
        # - On subsequent times (e.g. form submission), there will be a value
        #   in "target", so the default is irrelevant.
        # - This matters because we want "user_entry_node.description" to
        #   be correct.
        # - Actually, easier is just to make "target" a static display?
        #   No; if you use widget=TextInputWidget(readonly=True), there is no
        #   form value rendered.
        # - But it's hard to get the new value out of "target" at this point.
        # - Should we do that in validate()?
        # - No: on second rendering, after_bind() is called, and then
        #   validator() is called, but the visible form reflects changes made
        #   by after_bind() but NOT validator(); presumably Deform pulls the
        #   contents in between those two. Hmm.
        # - Particularly "hmm" as we don't have access to form data at the
        #   point of after_bind().
        # - The problem is probably that deform.field.Field.__init__ copies its
        #   schema.description. Yes, that's the problem.
        # - So: a third option: a display value (which we won't get back) as
        #   well as a hidden value that we will? No, no success.
        # - Or a fourth: something whose "description" is a property, not a
        #   str? No -- when you copy a property, you copy the value not the
        #   function.
        # - Fifthly: a new DangerValidationForm that rewrites its field
        #   descriptions after validation. That works!
        target_value = ''.join(random.choice(self.allowed_chars)
                               for i in range(self.length))
        target_node.default = target_value
        # Set the description:
        if DEBUG_DANGER_VALIDATION:
            log.debug("after_bind: setting description to {!r}", target_value)
        self.set_description(target_value)
        # ... may be overridden immediately by validator() if this is NOT the
        #     first rendering

    def validator(self, node: SchemaNode, value: Any) -> None:
        user_entry_value = value['user_entry']
        target_value = value['target']
        # Set the description:
        if DEBUG_DANGER_VALIDATION:
            log.debug("validator: setting description to {!r}", target_value)
        self.set_description(target_value)
        # arse!
        value['display_target'] = target_value
        # Check the value
        if user_entry_value != target_value:
            raise Invalid(
                node,
                f"Not correctly validated "
                f"(user_entry_value={user_entry_value!r}, "
                f"target_value={target_value!r}")

    def set_description(self, target_value: str) -> None:
        user_entry_node = get_child_node(self, "user_entry")
        prefix = "Please enter the following: "
        user_entry_node.description = prefix + target_value
        if DEBUG_DANGER_VALIDATION:
            log.debug("user_entry_node.description: {!r}",
                      user_entry_node.description)



#!/usr/bin/env python
# cardinal_pythonlib/maths_py.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Miscellaneous mathematical functions in pure Python.**

"""

import math
import sys
from typing import Optional, Sequence, Union

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Mean
# =============================================================================

def mean(values: Sequence[Union[int, float, None]]) -> Optional[float]:
    """
    Returns the mean of a list of numbers.

    Args:
        values: values to mean, ignoring any values that are ``None``

    Returns:
        the mean, or ``None`` if :math:`n = 0`

    """
    total = 0.0  # starting with "0.0" causes automatic conversion to float
    n = 0
    for x in values:
        if x is not None:
            total += x
            n += 1
    return total / n if n > 0 else None


# =============================================================================
# logit
# =============================================================================

def safe_logit(p: Union[float, int]) -> Optional[float]:
    r"""
    Returns the logit (log odds) of its input probability

    .. math::

        \alpha = logit(p) = log(x / (1 - x))

    Args:
        p: :math:`p`

    Returns:
        :math:`\alpha`, or ``None`` if ``x`` is not in the range [0, 1].

    """
    if p > 1 or p < 0:
        return None  # can't take log of negative number
    if p == 1:
        return float("inf")
    if p == 0:
        return float("-inf")
    return math.log(p / (1 - p))


# =============================================================================
# Rounding
# =============================================================================

def normal_round_float(x: float, dp: int = 0) -> float:
    """
    Hmpf. Shouldn't need to have to implement this, but...

    Conventional rounding to integer via the "round half away from zero"
    method, e.g.

    .. code-block:: none

        1.1 -> 1
        1.5 -> 2
        1.6 -> 2
        2.0 -> 2

        -1.6 -> -2
        etc.

    ... or the equivalent for a certain number of decimal places.

    Note that round() implements "banker's rounding", which is never what
    we want:
    - https://stackoverflow.com/questions/33019698/how-to-properly-round-up-half-float-numbers-in-python  # noqa
    """
    if not math.isfinite(x):
        return x
    factor = pow(10, dp)
    x = x * factor
    if x >= 0:
        x = math.floor(x + 0.5)
    else:
        x = math.ceil(x - 0.5)
    x = x / factor
    return x


def normal_round_int(x: float) -> int:
    """
    Version of :func:`normal_round_float` but guaranteed to return an `int`.
    """
    if not math.isfinite(x):
        raise ValueError("Input to normal_round_int() is not finite")
    if x >= 0:
        # noinspection PyTypeChecker
        return math.floor(x + 0.5)
    else:
        # noinspection PyTypeChecker
        return math.ceil(x - 0.5)


def round_sf(x: float, n: int = 2) -> float:
    """
    Round to a certain number of significant figures.

    As per https://code.activestate.com/lists/python-tutor/70739/, linked to
    from
    https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python

    Args:
        x: quantity to round
        n: number of significant figures

    Returns:
        float: x, rounded to n significant figures

    This does proper rounding:

    .. code-block:: none

        round_sf(0.55, 1)  # 0.6
        round_sf(0.549, 1)  # 0.5
        round_sf(-0.55, 1)  # -0.6
        round_sf(-0.549, 1)  # -0.5

        round_sf(0.000123456, 3)  # 0.000123
        round_sf(1234567890000, 3)  # 1230000000000
        round_sf(9876543210000, 3)  # 9880000000000

    """  # noqa
    y = abs(x)
    if y <= sys.float_info.min:
        return 0.0
    return round(x, int(n - math.ceil(math.log10(y))))


# =============================================================================
# Addition, permutation
# =============================================================================

def sum_of_integers_in_inclusive_range(a: int, b: int) -> int:
    """
    Returns the sum of all integers in the range ``[a, b]``, i.e. from ``a`` to
    ``b`` inclusive.

    See

    - https://math.stackexchange.com/questions/1842152/finding-the-sum-of-numbers-between-any-two-given-numbers
    """  # noqa
    return int((b - a + 1) * (a + b) / 2)


def n_permutations(n: int, k: int) -> int:
    """
    Returns the number of permutations of length ``k`` from a list of length
    ``n``.

    See https://en.wikipedia.org/wiki/Permutation#k-permutations_of_n.
    """
    assert n > 0 and 0 < k <= n
    return int(math.factorial(n) / math.factorial(n - k))



#!/usr/bin/env python
# cardinal_pythonlib/django/files.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Helper functions for managing disk files via Django.**

"""

import os
from typing import Any, Iterable


# =============================================================================
# Making FileFields own their files (i.e. delete them afterwards)
# =============================================================================

# https://stackoverflow.com/questions/16041232/django-delete-filefield
# These two auto-delete files from filesystem when they are unneeded:
# ... with a bit of modification to make them generic (RNC)
# Attach them with signals; see e.g. Study model.
def auto_delete_files_on_instance_delete(instance: Any,
                                         fieldnames: Iterable[str]) -> None:
    """
    Deletes files from filesystem when object is deleted.
    """
    for fieldname in fieldnames:
        filefield = getattr(instance, fieldname, None)
        if filefield:
            if os.path.isfile(filefield.path):
                os.remove(filefield.path)


# BugfixCT = TypeVar('BugfixCT', covariant=True, bound=type)
# # ... as per typing.CT (Python 3.5) or typing.CT_co (Python 3.6)
#
#
# class BugfixType(Generic[BugfixCT], extra=type):
#     """
#     Type checker note: using Type[Model] results in the error:
#     TypeError: descriptor '__subclasses__' of 'type' object needs an argument
#     ... which is the bug https://github.com/python/typing/issues/266
#
#     Using this fixes it.
#     The difference is that BugfixType doesn't inherit from type.
#
#     EXCEPT that the calling type checker from PyCharm doesn't know that it
#     can map Type[X] to BugfixType[X], so this is entirely unhelpful.
#     """


def auto_delete_files_on_instance_change(
        instance: Any,
        fieldnames: Iterable[str],
        model_class) -> None:
    """
    Deletes files from filesystem when object is changed.

    model_class: ``Type[Model]``

    ... only the type checker in Py3.5 is broken; v.s.
    """

    if not instance.pk:
        return  # instance not yet saved in database
    # noinspection PyUnresolvedReferences
    try:
        # noinspection PyUnresolvedReferences
        old_instance = model_class.objects.get(pk=instance.pk)
    except model_class.DoesNotExist:
        return  # old version gone from database entirely
    for fieldname in fieldnames:
        old_filefield = getattr(old_instance, fieldname, None)
        if not old_filefield:
            continue
        new_filefield = getattr(instance, fieldname, None)
        if old_filefield != new_filefield:
            if os.path.isfile(old_filefield.path):
                os.remove(old_filefield.path)



#!/usr/bin/env python

"""
tools/reformat_source.py

===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Clean up source code.**

"""

import logging
from os import walk
from os.path import join, splitext
from sys import stdout
from typing import List, TextIO

from cardinal_pythonlib.fileops import relative_filename_within_dir
from cardinal_pythonlib.logs import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

TRANSITION = "==============================================================================="  # noqa
CORRECT_SHEBANG = "#!/usr/bin/env python"
RST_COMMENT_LINE = ".."
SHEBANG_START = "#!"
TRIPLE_DOUBLEQUOTE = '"""'
RAW_TRIPLE_DOUBLEQUOTE = 'r"""'
BLANK = ""
MISSING_RST_TITLE = "**Missing title.**"

CR = "\r"
LF = "\n"
NL = LF
SPACE = " "
TAB = "\t"
HASH = "#"
HASH_SPACE = "# "
PYTHON_EXTENSION = ".py"


# =============================================================================
# PythonProcessor
# =============================================================================

class PythonProcessor(object):
    """
    Class to read a Python source file and reformat its shebang/docstring etc.
    """

    def __init__(self, full_path: str, top_dir: str,
                 correct_copyright_lines: List[str]) -> None:
        """

        Args:
            full_path:
                full path to source file
            top_dir:
                directory from which we calculate a relative filename to be
                shown
            correct_copyright_lines:
                list of lines (without newlines) representing the copyright
                docstring block, including the transition lines of equals
                symbols
        """
        self.full_path = full_path
        self.advertised_filename = relative_filename_within_dir(
            full_path, top_dir)
        self.correct_copyright_lines = correct_copyright_lines
        self.needs_rewriting = False
        self.source_lines = []  # type: List[str]
        self.dest_lines = []  # type: List[str]
        self._read_source()
        self._create_dest()

    def _read_source(self) -> None:
        """
        Reads the source file.
        """
        with open(self.full_path, "rt") as f:
            for linenum, line_with_nl in enumerate(f.readlines(), start=1):
                line_without_newline = (
                    line_with_nl[:-1] if line_with_nl.endswith(NL)
                    else line_with_nl
                )
                if TAB in line_without_newline:
                    self._warn(f"Tab character at line {linenum}")
                if CR in line_without_newline:
                    self._warn(f"Carriage return character at line {linenum} "
                               f"(Windows CR+LF endings?)")
                self.source_lines.append(line_without_newline)

    def _create_dest(self) -> None:
        """
        Creates an internal representation of the destination file.

        This is where the thinking happens
        """
        in_body = False
        in_docstring = False
        in_copyright = False
        copyright_done = False
        docstring_done = False
        swallow_blanks_and_filename_in_docstring = False
        for linenum, sl in enumerate(self.source_lines, start=1):
            dl = sl

            if dl.endswith(SPACE):
                self._debug(f"Line {linenum} ends in whitespace")
                dl = dl.rstrip()

            if not in_body:

                if linenum == 1:
                    # Shebang
                    if not dl.startswith(SHEBANG_START):
                        self._warn(f"File does not start with shebang; "
                                   f"first line was {dl!r}")
                        self._too_risky()
                        return
                    if dl != CORRECT_SHEBANG:
                        self._debug(f"Rewriting shebang; was {dl!r}")
                    dl = CORRECT_SHEBANG

                if (linenum == 2 and dl.startswith(HASH_SPACE) and
                        dl.endswith(PYTHON_EXTENSION)):
                    self._debug(
                        f"Removing filename comment: {dl!r}")
                    dl = None

                elif TRIPLE_DOUBLEQUOTE in dl:
                    if (not dl.startswith(TRIPLE_DOUBLEQUOTE) and
                            not dl.startswith(RAW_TRIPLE_DOUBLEQUOTE)):
                        self._warn(
                            "Triple-quote not at start of line, as follows")
                        self._debug_line(linenum, dl)
                        self._too_risky()
                        return
                    if in_docstring:  # docstring finishing
                        in_docstring = False
                        docstring_done = True
                        in_body = True
                        # ... and keep dl, so we write the end of the
                        # docstring, potentially with e.g. "# noqa" on the end
                    elif not docstring_done:  # docstring starting
                        in_docstring = True
                        # self._critical("adding our new docstring")
                        # Write our new docstring's start
                        tdq = ""  # stops linter moaning
                        if dl.startswith(TRIPLE_DOUBLEQUOTE):
                            tdq = TRIPLE_DOUBLEQUOTE
                        elif dl.startswith(RAW_TRIPLE_DOUBLEQUOTE):
                            tdq = RAW_TRIPLE_DOUBLEQUOTE
                        else:
                            assert "Bug!"
                        self.dest_lines.append(tdq)
                        self.dest_lines.append(self.advertised_filename)
                        self.dest_lines.append(BLANK)
                        self.dest_lines.extend(self.correct_copyright_lines)
                        self.dest_lines.append(BLANK)
                        swallow_blanks_and_filename_in_docstring = True
                        if dl == tdq:
                            dl = None  # don't write another triple-quote line
                        else:
                            dl = dl[len(tdq):]

                elif in_docstring:
                    # Reading within the source docstring

                    if dl == TRANSITION:
                        if in_copyright:  # copyright finishing
                            in_copyright = False
                            copyright_done = True
                            dl = None  # we've already replaced with our own
                        elif not copyright_done:
                            in_copyright = True
                            dl = None  # we've already replaced with our own

                    elif in_copyright:
                        dl = None  # we've already replaced with our own

                    elif dl == RST_COMMENT_LINE:
                        dl = None  # remove these

                    elif swallow_blanks_and_filename_in_docstring:
                        # self._debug_line(linenum, dl)
                        if dl == BLANK or dl == self.advertised_filename:
                            dl = None
                        elif copyright_done:
                            swallow_blanks_and_filename_in_docstring = False

                elif not dl.startswith(HASH) and not dl == BLANK:
                    in_body = True

                    if not docstring_done:
                        # The source file didn't have a docstring!
                        new_docstring_lines = [
                            BLANK,
                            TRIPLE_DOUBLEQUOTE,
                            self.advertised_filename,
                            BLANK,
                        ] + self.correct_copyright_lines + [
                            BLANK,
                            MISSING_RST_TITLE,
                            BLANK,
                            TRIPLE_DOUBLEQUOTE
                        ]
                        self._warn(f"File had no docstring; adding one. "
                                   f"Will need manual edit to add RST title. "
                                   f"Search for {MISSING_RST_TITLE!r}")
                        self.dest_lines[1:1] = new_docstring_lines

            if dl is not None:
                # self._debug_line(linenum, dl, "adding ")
                self.dest_lines.append(dl)

        self.needs_rewriting = self.dest_lines != self.source_lines

    @staticmethod
    def _debug_line(linenum: int, line: str, extramsg: str = "") -> None:
        """
        Writes a debugging report on a line.
        """
        log.critical("{}Line {}: {!r}", extramsg, linenum, line)

    def _logmsg(self, msg: str) -> str:
        """
        Formats a log message.
        """
        return f"{self.advertised_filename}: {msg}"

    def _critical(self, msg: str) -> None:
        """
        Shows a critical message.
        """
        log.critical(self._logmsg(msg))

    def _warn(self, msg: str) -> None:
        """
        Shows a warning.
        """
        log.warning(self._logmsg(msg))

    def _info(self, msg: str) -> None:
        """
        Shows an info message.
        """
        log.info(self._logmsg(msg))

    def _debug(self, msg: str) -> None:
        """
        Shows a debugging message.
        """
        log.debug(self._logmsg(msg))

    def _too_risky(self) -> None:
        """
        Shows a warning and sets this file as not for processing
        """
        self._warn("Don't know how to process file")
        self.needs_rewriting = False

    def show(self) -> None:
        """
        Writes the destination to stdout.
        """
        self._write(stdout)

    def rewrite_file(self) -> None:
        """
        Rewrites the source file.
        """
        if not self.needs_rewriting:
            return
        self._info("Rewriting file")
        with open(self.full_path, "w") as outfile:
            self._write(outfile)

    def _write(self, destination: TextIO) -> None:
        """
        Writes the converted output to a destination.
        """
        for line in self.dest_lines:
            destination.write(line + NL)


# =============================================================================
# Top-level functions
# =============================================================================

def reformat_python_docstrings(top_dirs: List[str],
                               correct_copyright_lines: List[str],
                               show_only: bool = True,
                               rewrite: bool = False,
                               process_only_filenum: int = None) -> None:
    """
    Walk a directory, finding Python files and rewriting them.

    Args:
        top_dirs: list of directories to descend into
        correct_copyright_lines:
            list of lines (without newlines) representing the copyright
            docstring block, including the transition lines of equals
            symbols
        show_only: show results (to stdout) only; don't rewrite
        rewrite: write the changes
        process_only_filenum: only process this file number (1-based index);
            for debugging only
    """
    filenum = 0
    for top_dir in top_dirs:
        for dirpath, dirnames, filenames in walk(top_dir):
            for filename in filenames:
                fullname = join(dirpath, filename)
                extension = splitext(filename)[1]
                if extension != PYTHON_EXTENSION:
                    # log.debug("Skipping non-Python file: {}", fullname)
                    continue

                filenum += 1

                if process_only_filenum and filenum != process_only_filenum:
                    continue

                log.info("Processing file {}: {}", filenum, fullname)
                proc = PythonProcessor(
                    full_path=fullname,
                    top_dir=top_dir,
                    correct_copyright_lines=correct_copyright_lines)
                if show_only:
                    proc.show()
                elif rewrite:
                    proc.rewrite_file()



#!/usr/bin/env python
# cardinal_pythonlib/pyramid/requests.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

Functions to operate on Pyramid requests.

"""

import gzip
import logging
from typing import Generator
import zlib

# noinspection PyUnresolvedReferences
from pyramid.request import Request
# noinspection PyUnresolvedReferences
from webob.headers import EnvironHeaders

try:
    # noinspection PyPackageRequirements
    import brotli  # pip install brotlipy
except ImportError:
    brotli = None

log = logging.getLogger(__name__)


# =============================================================================
# Encoding checks
# =============================================================================

HTTP_ACCEPT_ENCODING = "Accept-Encoding"
HTTP_CONTENT_ENCODING = "Content-Encoding"

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Encoding
BR_ENCODING = "br"
COMPRESS_ENCODING = "compress"
DEFLATE_ENCODING = "deflate"
GZIP_ENCODING = "gzip"
X_GZIP_ENCODING = "x-gzip"
IDENTITY_ENCODING = "identity"


def gen_accept_encoding_definitions(accept_encoding: str) \
        -> Generator[str, None, None]:
    """
    For a given HTTP ``Accept-Encoding`` field value, generate encoding
    definitions. An example might be:

    .. code-block:: python

        from cardinal_pythonlib.pyramid.compression import *
        accept_encoding = "br;q=1.0, gzip;q=0.8, *;q=0.1"
        print(list(gen_encoding_definitions(accept_encoding)))

    which gives

    .. code-block:: none

        ['br;q=1.0', 'gzip;q=0.8', '*;q=0.1']

    """
    for definition in accept_encoding.split(","):
        yield definition.strip()


def gen_accept_encodings(accept_encoding: str) \
        -> Generator[str, None, None]:
    """
    For a given HTTP ``Accept-Encoding`` field value, generate encodings.
    An example might be:

    .. code-block:: python

        from cardinal_pythonlib.pyramid.compression import *
        accept_encoding = "br;q=1.0, gzip;q=0.8, *;q=0.1"
        print(list(gen_encodings(accept_encoding)))

    which gives

    .. code-block:: none

        ['br', 'gzip', '*']
    """
    for definition in gen_accept_encoding_definitions(accept_encoding):
        yield definition.split(";")[0].strip()


def request_accepts_gzip(request: Request) -> bool:
    """
    Does the request specify an ``Accept-Encoding`` header that includes
    ``gzip``?

    Note:

    - Field names in HTTP headers are case-insensitive (e.g. "accept-encoding"
      is fine).
    - WebOb request headers are in a case-insensitive dictionary; see
      https://docs.pylonsproject.org/projects/webob/en/stable/. (Easily
      verified by altering the key being checked; it works.)
    - However, the value is a Python string so is case-sensitive.
    - But the HTTP standard doesn't say that field values are case-insensitive;
      see https://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2.
    - So we'll do a case-sensitive check for "gzip".
    - But there is also a bit of other syntax possible; see
      https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Encoding.
    """  # noqa
    headers = request.headers  # type: EnvironHeaders
    if HTTP_ACCEPT_ENCODING not in headers:
        return False
    accepted_encodings = headers[HTTP_ACCEPT_ENCODING]
    for encoding in gen_accept_encodings(accepted_encodings):
        if encoding == GZIP_ENCODING:
            return True
    return False


def gen_content_encodings(request: Request) -> Generator[str, None, None]:
    """
    Generates content encodings in the order they are specified -- that is, the
    order in which they were applied.
    """
    headers = request.headers  # type: EnvironHeaders
    if HTTP_CONTENT_ENCODING not in headers:
        return
    content_encoding = headers[HTTP_CONTENT_ENCODING]
    for encoding in content_encoding.split(","):
        yield encoding.strip()


def gen_content_encodings_reversed(request: Request) \
        -> Generator[str, None, None]:
    """
    Generates content encodings in reverse order -- that is, in the order
    required to reverse them.
    """
    for encoding in reversed(list(gen_content_encodings(request))):
        yield encoding


def decompress_request(request: Request) -> None:
    """
    Reverses anything specified in ``Content-Encoding``, modifying the request
    in place.
    """
    for encoding in gen_content_encodings_reversed(request):
        if encoding in [GZIP_ENCODING, X_GZIP_ENCODING]:
            request.body = gzip.decompress(request.body)
        elif encoding == IDENTITY_ENCODING:
            pass
        elif encoding == DEFLATE_ENCODING:
            request.body = zlib.decompress(request.body)
        elif encoding == BR_ENCODING:
            if brotli:
                # https://python-hyper.org/projects/brotlipy/en/latest/
                # noinspection PyUnresolvedReferences
                request.body = brotli.decompress(request.body)
            else:
                raise NotImplementedError(
                    f"Content-Encoding {encoding} not supported "
                    f"(brotlipy package not installed)")
        else:
            raise ValueError(f"Unknown Content-Encoding: {encoding}")
            # ... e.g. "compress"; LZW; patent expired; see
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Encoding  # noqa



#!/usr/bin/env python
# cardinal_pythonlib/rnc_web.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support for web scripts.**

"""


import base64
import binascii
import cgi
import configparser
import dateutil.parser
import dateutil.tz
import datetime
import html
import os
import re
import sys
from typing import (Any, Callable, Dict, Iterable, List, Optional,
                    Tuple, Union)

from cardinal_pythonlib.wsgi.constants import (
    TYPE_WSGI_APP_RESULT,
    TYPE_WSGI_START_RESPONSE,
    TYPE_WSGI_RESPONSE_HEADERS,
)
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)

WSGI_TUPLE_TYPE = Tuple[str, TYPE_WSGI_RESPONSE_HEADERS, bytes]
# ... contenttype, extraheaders, output

# =============================================================================
# Constants
# =============================================================================

_NEWLINE_REGEX = re.compile("\n", re.MULTILINE)
BASE64_PNG_URL_PREFIX = "data:image/png;base64,"
PNG_SIGNATURE_HEXSTRING = "89504E470D0A1A0A"
# ... http://en.wikipedia.org/wiki/Portable_Network_Graphics#Technical_details
PNG_SIGNATURE_HEX = binascii.unhexlify(PNG_SIGNATURE_HEXSTRING)
# ... bytes in Python 3; str in Python 2


# =============================================================================
# Misc
# =============================================================================

def print_utf8(s: str) -> None:
    """
    Writes a Unicode string to ``sys.stdout`` in UTF-8 encoding.
    """
    sys.stdout.buffer.write(s.encode('utf-8'))


def get_int_or_none(s: str) -> Optional[int]:
    """
    Returns the integer value of a string, or ``None`` if it's not convertible
    to an ``int``.
    """
    try:
        return int(s)
        # int(x) will return something of type long if it's a big number,
        # but happily
    except (TypeError, ValueError):
        return None


def get_float_or_none(s: str) -> Optional[float]:
    """
    Returns the float value of a string, or ``None`` if it's not convertible
    to a ``float``.
    """
    try:
        return float(s)
    except (TypeError, ValueError):
        return None


def is_1(s: str) -> bool:
    """
    ``True`` if the input is the string literal ``"1"``, otherwise ``False``.
    """
    return True if s == "1" else False


def number_to_dp(number: Optional[float],
                 dp: int,
                 default: Optional[str] = "",
                 en_dash_for_minus: bool = True) -> str:
    """
    Format number to ``dp`` decimal places, optionally using a UTF-8 en dash
    for minus signs.
    """
    if number is None:
        return default
    if number == float("inf"):
        return u"∞"
    if number == float("-inf"):
        s = u"-∞"
    else:
        s = f"{number:.{dp}f}"
    if en_dash_for_minus:
        s = s.replace("-", u"–")  # hyphen becomes en dash for minus sign
    return s


# =============================================================================
# CGI
# =============================================================================

def debug_form_contents(form: cgi.FieldStorage,
                        to_stderr: bool = True,
                        to_logger: bool = False) -> None:
    """
    Writes the keys and values of a CGI form to ``stderr``.
    """
    for k in form.keys():
        text = f"{k} = {form.getvalue(k)}"
        if to_stderr:
            sys.stderr.write(text)
        if to_logger:
            log.info(text)
    # But note also: cgi.print_form(form)


def cgi_method_is_post(environ: Dict[str, str]) -> bool:
    """
    Determines if the CGI method was ``POST``, given the CGI environment.
    """
    method = environ.get("REQUEST_METHOD", None)
    if not method:
        return False
    return method.upper() == "POST"


def get_cgi_parameter_str(form: cgi.FieldStorage,
                          key: str,
                          default: str = None) -> str:
    """
    Extracts a string parameter from a CGI form.
    Note: ``key`` is CASE-SENSITIVE.
    """
    paramlist = form.getlist(key)
    if len(paramlist) == 0:
        return default
    return paramlist[0]


def get_cgi_parameter_str_or_none(form: cgi.FieldStorage,
                                  key: str) -> Optional[str]:
    """
    Extracts a string parameter from a CGI form, or ``None`` if the key doesn't
    exist or the string is zero-length.
    """
    s = get_cgi_parameter_str(form, key)
    if s is None or len(s) == 0:
        return None
    return s


def get_cgi_parameter_list(form: cgi.FieldStorage, key: str) -> List[str]:
    """
    Extracts a list of values, all with the same key, from a CGI form.
    """
    return form.getlist(key)


def get_cgi_parameter_bool(form: cgi.FieldStorage, key: str) -> bool:
    """
    Extracts a boolean parameter from a CGI form, on the assumption that
    ``"1"`` is ``True`` and everything else is ``False``.
    """
    return is_1(get_cgi_parameter_str(form, key))


def get_cgi_parameter_bool_or_default(form: cgi.FieldStorage,
                                      key: str,
                                      default: bool = None) -> Optional[bool]:
    """
    Extracts a boolean parameter from a CGI form (``"1"`` = ``True``,
    other string = ``False``, absent/zero-length string = default value).
    """
    s = get_cgi_parameter_str(form, key)
    if s is None or len(s) == 0:
        return default
    return is_1(s)


def get_cgi_parameter_bool_or_none(form: cgi.FieldStorage,
                                   key: str) -> Optional[bool]:
    """
    Extracts a boolean parameter from a CGI form (``"1"`` = ``True``,
    other string = False, absent/zero-length string = ``None``).
    """
    return get_cgi_parameter_bool_or_default(form, key, default=None)


def get_cgi_parameter_int(form: cgi.FieldStorage, key: str) -> Optional[int]:
    """
    Extracts an integer parameter from a CGI form, or ``None`` if the key is
    absent or the string value is not convertible to ``int``.
    """
    return get_int_or_none(get_cgi_parameter_str(form, key))


def get_cgi_parameter_float(form: cgi.FieldStorage,
                            key: str) -> Optional[float]:
    """
    Extracts a float parameter from a CGI form, or None if the key is
    absent or the string value is not convertible to ``float``.
    """
    return get_float_or_none(get_cgi_parameter_str(form, key))


def get_cgi_parameter_datetime(form: cgi.FieldStorage,
                               key: str) -> Optional[datetime.datetime]:
    """
    Extracts a date/time parameter from a CGI form. Applies the LOCAL
    timezone if none specified.
    """
    try:
        s = get_cgi_parameter_str(form, key)
        if not s:
            # if you dateutil.parser.parse() an empty string,
            # you get today's date
            return None
        d = dateutil.parser.parse(s)
        if d.tzinfo is None:  # as it will be
            d = d.replace(tzinfo=dateutil.tz.tzlocal())
        return d
    except ValueError:
        return None


def get_cgi_parameter_file(form: cgi.FieldStorage,
                           key: str) -> Optional[bytes]:
    """
    Extracts a file's contents from a "file" input in a CGI form, or None
    if no such file was uploaded.
    """
    (filename, filecontents) = get_cgi_parameter_filename_and_file(form, key)
    return filecontents


def get_cgi_parameter_filename_and_file(form: cgi.FieldStorage, key: str) \
        -> Tuple[Optional[str], Optional[bytes]]:
    """
    Extracts a file's name and contents from a "file" input in a CGI form.
    Returns ``(name, contents)``, or ``(None, None)`` if no such file was
    uploaded.
    """
    if not (key in form):
        log.warning('get_cgi_parameter_file: form has no key {}', key)
        return None, None
    fileitem = form[key]  # a nested FieldStorage instance; see
    # http://docs.python.org/2/library/cgi.html#using-the-cgi-module
    if isinstance(fileitem, cgi.MiniFieldStorage):
        log.warning('get_cgi_parameter_file: MiniFieldStorage found - did you '
                    'forget to set enctype="multipart/form-data" in '
                    'your form?')
        return None, None
    if not isinstance(fileitem, cgi.FieldStorage):
        log.warning('get_cgi_parameter_file: no FieldStorage instance with '
                    'key {} found', key)
        return None, None
    if fileitem.filename and fileitem.file:  # can check "file" or "filename"
        return fileitem.filename, fileitem.file.read()
        # as per
        # http://upsilon.cc/~zack/teaching/0607/techweb/02-python-cgi.pdf
        # Alternative:
        # return get_cgi_parameter_str(form, key) # contents of the file
    # Otherwise, information about problems:
    if not fileitem.file:
        log.warning('get_cgi_parameter_file: fileitem has no file')
    elif not fileitem.filename:
        log.warning('get_cgi_parameter_file: fileitem has no filename')
    else:
        log.warning('get_cgi_parameter_file: unknown failure reason')
    return None, None

    # "If a field represents an uploaded file, accessing the value
    # via the value attribute or the getvalue() method reads the
    # entire file in memory as a string. This may not be what you
    # want. You can test for an uploaded file by testing either
    # the filename attribute or the file attribute. You can then
    # read the data at leisure from the file attribute:"


def cgi_parameter_exists(form: cgi.FieldStorage, key: str) -> bool:
    """
    Does a CGI form contain the key?
    """
    s = get_cgi_parameter_str(form, key)
    return s is not None


def checkbox_checked(b: Any) -> str:
    """
    Returns ``' checked="checked"'`` if ``b`` is true; otherwise ``''``.

    Use this code to fill the ``{}`` in e.g.:

    .. code-block:: none

        <label>
            <input type="checkbox" name="myfield" value="1"{}>
            This will be pre-ticked if you insert " checked" where the braces
            are. The newer, more stringent requirement is ' checked="checked"'.
        </label>
    """
    return ' checked="checked"' if b else ''


def option_selected(variable: Any, testvalue: Any) -> str:
    """
    Returns ``' selected="selected"'`` if ``variable == testvalue`` else
    ``''``; for use with HTML select options.
    """
    return ' selected="selected"' if variable == testvalue else ''


# =============================================================================
# Environment
# =============================================================================

def getenv_escaped(key: str, default: str = None) -> Optional[str]:
    """
    Returns an environment variable's value, CGI-escaped, or ``None``.
    """
    value = os.getenv(key, default)
    return html.escape(value) if value is not None else None


def getconfigvar_escaped(config: configparser.ConfigParser,
                         section: str,
                         key: str) -> Optional[str]:
    """
    Returns a CGI-escaped version of the value read from an INI file using
    :class:`ConfigParser`, or ``None``.
    """
    value = config.get(section, key)
    # noinspection PyDeprecation
    return html.escape(value) if value is not None else None


def get_cgi_fieldstorage_from_wsgi_env(
        env: Dict[str, str],
        include_query_string: bool = True) -> cgi.FieldStorage:
    """
    Returns a :class:`cgi.FieldStorage` object from the WSGI environment.
    """
    # https://stackoverflow.com/questions/530526/accessing-post-data-from-wsgi
    post_env = env.copy()
    if not include_query_string:
        post_env['QUERY_STRING'] = ''
    form = cgi.FieldStorage(
        fp=env['wsgi.input'],
        environ=post_env,
        keep_blank_values=True
    )
    return form


# =============================================================================
# Blobs, pictures...
# =============================================================================

def is_valid_png(blob: Optional[bytes]) -> bool:
    """
    Does a blob have a valid PNG signature?
    """
    if not blob:
        return False
    return blob[:8] == PNG_SIGNATURE_HEX


def get_png_data_url(blob: Optional[bytes]) -> str:
    """
    Converts a PNG blob into a local URL encapsulating the PNG.
    """
    return BASE64_PNG_URL_PREFIX + base64.b64encode(blob).decode('ascii')


def get_png_img_html(blob: Union[bytes, memoryview],
                     extra_html_class: str = None) -> str:
    """
    Converts a PNG blob to an HTML IMG tag with embedded data.
    """
    return '<img {}src="{}" />'.format(
        f'class="{extra_html_class}" ' if extra_html_class else "",
        get_png_data_url(blob)
    )


# =============================================================================
# HTTP results
# =============================================================================

# Also, filenames:
#   https://stackoverflow.com/questions/151079
#   http://greenbytes.de/tech/tc2231/#inlwithasciifilenamepdf

def pdf_result(pdf_binary: bytes,
               extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None,
               filename: str = None) -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for a PDF.
    """
    extraheaders = extraheaders or []
    if filename:
        extraheaders.append(
            ('content-disposition', f'inline; filename="{filename}"')
        )
    contenttype = 'application/pdf'
    if filename:
        contenttype += f'; filename="{filename}"'
    # log.debug("type(pdf_binary): {}", type(pdf_binary))
    return contenttype, extraheaders, pdf_binary


def zip_result(zip_binary: bytes,
               extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None,
               filename: str = None) -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for a ZIP.
    """
    extraheaders = extraheaders or []
    if filename:
        extraheaders.append(
            ('content-disposition', f'inline; filename="{filename}"')
        )
    contenttype = 'application/zip'
    if filename:
        contenttype += f'; filename="{filename}"'
    return contenttype, extraheaders, zip_binary


# noinspection PyShadowingNames
def html_result(html: str,
                extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None) \
        -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for UTF-8 HTML.
    """
    extraheaders = extraheaders or []
    return 'text/html; charset=utf-8', extraheaders, html.encode("utf-8")


def xml_result(xml: str,
               extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None) \
        -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for UTF-8 XML.
    """
    extraheaders = extraheaders or []
    return 'text/xml; charset=utf-8', extraheaders, xml.encode("utf-8")


def text_result(text: str,
                extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None,
                filename: str = None) -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for UTF-8 text.
    """
    extraheaders = extraheaders or []
    if filename:
        extraheaders.append(
            ('content-disposition', f'inline; filename="{filename}"')
        )
    contenttype = 'text/plain; charset=utf-8'
    if filename:
        contenttype += f'; filename="{filename}"'
    return contenttype, extraheaders, text.encode("utf-8")


def tsv_result(text: str,
               extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None,
               filename: str = None) -> WSGI_TUPLE_TYPE:
    """
    Returns ``(contenttype, extraheaders, data)`` tuple for UTF-8 TSV.
    """
    extraheaders = extraheaders or []
    if filename:
        extraheaders.append(
            ('content-disposition', f'inline; filename="{filename}"')
        )
    contenttype = 'text/tab-separated-values; charset=utf-8'
    if filename:
        contenttype += f'; filename="{filename}"'
    return contenttype, extraheaders, text.encode("utf-8")


# =============================================================================
# CGI
# =============================================================================

def print_result_for_plain_cgi_script_from_tuple(
        contenttype_headers_content: WSGI_TUPLE_TYPE,
        status: str = '200 OK') -> None:
    """
    Writes HTTP result to stdout.

    Args:
        contenttype_headers_content:
            the tuple ``(contenttype, extraheaders, data)``
        status:
            HTTP status message (default ``"200 OK``)
    """
    contenttype, headers, content = contenttype_headers_content
    print_result_for_plain_cgi_script(contenttype, headers, content, status)


def print_result_for_plain_cgi_script(contenttype: str,
                                      headers: TYPE_WSGI_RESPONSE_HEADERS,
                                      content: bytes,
                                      status: str = '200 OK') -> None:
    """
    Writes HTTP request result to stdout.
    """
    headers = [
        ("Status", status),
        ("Content-Type", contenttype),
        ("Content-Length", str(len(content))),
    ] + headers
    sys.stdout.write("\n".join([h[0] + ": " + h[1] for h in headers]) + "\n\n")
    sys.stdout.buffer.write(content)


# =============================================================================
# WSGI
# =============================================================================

def wsgi_simple_responder(
        result: Union[str, bytes],
        handler: Callable[[Union[str, bytes]], WSGI_TUPLE_TYPE],
        start_response: TYPE_WSGI_START_RESPONSE,
        status: str = '200 OK',
        extraheaders: TYPE_WSGI_RESPONSE_HEADERS = None) \
        -> TYPE_WSGI_APP_RESULT:
    """
    Simple WSGI app.

    Args:
        result: the data to be processed by ``handler``
        handler: a function returning a ``(contenttype, extraheaders, data)``
            tuple, e.g. ``text_result``, ``html_result``
        start_response: standard WSGI ``start_response`` function
        status: status code (default ``"200 OK"``)
        extraheaders: optional extra HTTP headers

    Returns:
        WSGI application result

    """
    extraheaders = extraheaders or []
    (contenttype, extraheaders2, output) = handler(result)
    response_headers = [('Content-Type', contenttype),
                        ('Content-Length', str(len(output)))]
    response_headers.extend(extraheaders)
    if extraheaders2 is not None:
        response_headers.extend(extraheaders2)
    # noinspection PyArgumentList
    start_response(status, response_headers)
    return [output]


# =============================================================================
# HTML
# =============================================================================

def webify(v: Any, preserve_newlines: bool = True) -> str:
    """
    Converts a value into an HTML-safe ``str`` (formerly, in Python 2:
    ``unicode``).

    Converts value ``v`` to a string; escapes it to be safe in HTML
    format (escaping ampersands, replacing newlines with ``<br>``, etc.).
    Returns ``""`` for blank input.
    """
    nl = "<br>" if preserve_newlines else " "
    if v is None:
        return ""
    if not isinstance(v, str):
        v = str(v)
    return html.escape(v).replace("\n", nl).replace("\\n", nl)


def websafe(value: str) -> str:
    """
    Makes a string safe for inclusion in ASCII-encoded HTML.
    """
    return html.escape(value).encode('ascii', 'xmlcharrefreplace')
    # https://stackoverflow.com/questions/1061697


def replace_nl_with_html_br(string: str) -> str:
    """
    Replaces newlines with ``<br>``.
    """
    return _NEWLINE_REGEX.sub("<br>", string)


def bold_if_not_blank(x: Optional[str]) -> str:
    """
    HTML-emboldens content, unless blank.
    """
    if x is None:
        return f"{x}"
    return f"<b>{x}</b>"


def make_urls_hyperlinks(text: str) -> str:
    # noinspection HttpUrlsUsage
    """
    Adds hyperlinks to text that appears to contain URLs.

    See

    - https://stackoverflow.com/questions/1071191

      - ... except that double-replaces everything; e.g. try with
        ``text = "me@somewhere.com me@somewhere.com"``

    - http://stackp.online.fr/?p=19
    """
    find_url = r'''
        (?x)(              # verbose identify URLs within text
        (http|ftp|gopher)  # make sure we find a resource type
        ://                # ...needs to be followed by colon-slash-slash
        (\w+[:.]?){2,}     # at least two domain groups, e.g. (gnosis.)(cx)
        (/?|               # could be just the domain name (maybe w/ slash)
        [^ \n\r"]+         # or stuff then space, newline, tab, quote
        [\w/])             # resource name ends in alphanumeric or slash
        (?=[\s\.,>)'"\]])  # assert: followed by white or clause ending
        )                  # end of match group
    '''
    replace_url = r'<a href="\1">\1</a>'
    find_email = re.compile(r'([.\w\-]+@(\w[\w\-]+\.)+[\w\-]+)')
    # '.' doesn't need escaping inside square brackets
    # https://stackoverflow.com/questions/10397968/escape-dot-in-a-regex-range
    replace_email = r'<a href="mailto:\1">\1</a>'
    text = re.sub(find_url, replace_url, text)
    text = re.sub(find_email, replace_email, text)
    return text


# noinspection PyShadowingNames
def html_table_from_query(rows: Iterable[Iterable[Optional[str]]],
                          descriptions: Iterable[Optional[str]]) -> str:
    """
    Converts rows from an SQL query result to an HTML table.
    Suitable for processing output from the defunct function
    ``rnc_db.fetchall_with_fieldnames(sql)``.
    """
    html = u"<table>\n"

    # Header row
    html += u"<tr>"
    for x in descriptions:
        if x is None:
            x = u""
        html += f"<th>{webify(x)}</th>"
    html += u"</tr>\n"

    # Data rows
    for row in rows:
        html += u"<tr>"
        for x in row:
            if x is None:
                x = u""
            html += f"<td>{webify(x)}</td>"
        html += u"<tr>\n"

    html += u"</table>\n"
    return html



#!/usr/bin/env python
# cardinal_pythonlib/tools/merge_csv.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Command-line tool to merge multiple comma-separated value (CSV) or
tab-separated value (TSV) files, as long as they don't have incompatible
headers.**

"""

import argparse
import csv
import logging
import sys
from typing import List, TextIO

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


def merge_csv(filenames: List[str],
              outfile: TextIO = sys.stdout,
              input_dialect: str = 'excel',
              output_dialect: str = 'excel',
              debug: bool = False,
              headers: bool = True) -> None:
    """
    Amalgamate multiple CSV/TSV/similar files into one.

    Args:
        filenames: list of filenames to process
        outfile: file-like object to write output to
        input_dialect: dialect of input files, as passed to ``csv.reader``
        output_dialect: dialect to write, as passed to ``csv.writer``
        debug: be verbose?
        headers: do the files have header lines?
    """
    writer = csv.writer(outfile, dialect=output_dialect)
    written_header = False
    header_items = []  # type: List[str]
    for filename in filenames:
        log.info("Processing file " + repr(filename))
        with open(filename, 'r') as f:
            reader = csv.reader(f, dialect=input_dialect)
            if headers:
                if not written_header:
                    header_items = next(reader)
                    if debug:
                        log.debug("Header row: {!r}", header_items)
                    writer.writerow(header_items)
                    written_header = True
                else:
                    new_headers = next(reader)
                    if new_headers != header_items:
                        raise ValueError(
                            f"Header line in file {filename!r} doesn't match "
                            f"- it was {new_headers!r} but previous was "
                            f"{header_items!r}")
                    if debug:
                        log.debug("Header row matches previous")
            else:
                if debug:
                    log.debug("No headers in use")
            for row in reader:
                if debug:
                    log.debug("Data row: {!r}", row)
                writer.writerow(row)


def main():
    """
    Command-line processor. See ``--help`` for details.
    """
    main_only_quicksetup_rootlogger()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filenames",
        nargs="+",
        help="Names of CSV/TSV files to merge"
    )
    parser.add_argument(
        "--outfile",
        default="-",
        help="Specify an output filename. If omitted or '-', stdout is used.",
    )
    parser.add_argument(
        "--inputdialect",
        default="excel",
        help="The input files' CSV/TSV dialect. Default: %(default)s.",
        choices=csv.list_dialects(),
    )
    parser.add_argument(
        "--outputdialect",
        default="excel",
        help="The output file's CSV/TSV dialect. Default: %(default)s.",
        choices=csv.list_dialects(),
    )
    parser.add_argument(
        "--noheaders",
        action="store_true",
        help="By default, files are assumed to have column headers. "
             "Specify this option to assume no headers.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Verbose debugging output.",
    )
    progargs = parser.parse_args()

    kwargs = {
        "filenames": progargs.filenames,
        "input_dialect": progargs.inputdialect,
        "output_dialect": progargs.outputdialect,
        "debug": progargs.debug,
        "headers": not progargs.noheaders,
    }
    if progargs.outfile == '-':
        log.info("Writing to stdout")
        merge_csv(outfile=sys.stdout, **kwargs)
    else:
        log.info("Writing to " + repr(progargs.outfile))
        with open(progargs.outfile, 'w') as outfile:
            # noinspection PyTypeChecker
            merge_csv(outfile=outfile, **kwargs)


if __name__ == '__main__':
    main()
